from __future__ import annotations

from pathlib import Path
from typing import Any, Optional

from chik_rs import G2Element

from chik._tests.cmds.cmd_test_utils import TestRpcClients, TestWalletRpcClient, logType, run_cli_command_and_assert
from chik._tests.cmds.wallet.test_consts import FINGERPRINT, FINGERPRINT_ARG, STD_TX, STD_UTX, get_bytes32
from chik.rpc.wallet_request_types import (
    NFTAddURIResponse,
    NFTMintNFTResponse,
    NFTSetNFTDIDResponse,
    NFTTransferNFTResponse,
)
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.types.signing_mode import SigningMode
from chik.util.bech32m import encode_puzzle_hash
from chik.util.ints import uint8, uint16, uint32, uint64
from chik.wallet.conditions import ConditionValidTimes
from chik.wallet.nft_wallet.nft_info import NFTInfo
from chik.wallet.util.tx_config import DEFAULT_TX_CONFIG, TXConfig
from chik.wallet.wallet_spend_bundle import WalletSpendBundle

test_condition_valid_times: ConditionValidTimes = ConditionValidTimes(min_time=uint64(100), max_time=uint64(150))

# NFT Commands


def test_nft_create(capsys: object, get_test_cli_clients: tuple[TestRpcClients, Path]) -> None:
    test_rpc_clients, root_dir = get_test_cli_clients

    # set RPC Client
    class NFTCreateRpcClient(TestWalletRpcClient):
        async def create_new_nft_wallet(self, did_id: str, name: Optional[str] = None) -> dict[str, Any]:
            self.add_to_log("create_new_nft_wallet", (did_id, name))
            return {"wallet_id": 4}

    inst_rpc_client = NFTCreateRpcClient()
    did_id = encode_puzzle_hash(get_bytes32(2), "did:chik:")
    test_rpc_clients.wallet_rpc_client = inst_rpc_client
    command_args = ["wallet", "nft", "create", FINGERPRINT_ARG, "-ntest", "--did-id", did_id]
    # these are various things that should be in the output
    assert_list = [f"Successfully created an NFT wallet with id 4 on key {FINGERPRINT}"]
    run_cli_command_and_assert(capsys, root_dir, command_args, assert_list)
    expected_calls: logType = {
        "create_new_nft_wallet": [(did_id, "test")],
    }
    test_rpc_clients.wallet_rpc_client.check_log(expected_calls)


def test_nft_sign_message(capsys: object, get_test_cli_clients: tuple[TestRpcClients, Path]) -> None:
    test_rpc_clients, root_dir = get_test_cli_clients

    # set RPC Client

    inst_rpc_client = TestWalletRpcClient()
    test_rpc_clients.wallet_rpc_client = inst_rpc_client
    did_id = encode_puzzle_hash(get_bytes32(1), "nft")
    message = b"hello nft world!!"
    command_args = ["wallet", "did", "sign_message", FINGERPRINT_ARG, f"-m{message.hex()}"]
    # these are various things that should be in the output
    assert_list = [
        f"Message: {message.hex()}",
        f"Public Key: {bytes([4] * 48).hex()}",
        f"Signature: {bytes([7] * 576).hex()}",
        f"Signing Mode: {SigningMode.CHIP_0002.value}",
    ]
    run_cli_command_and_assert(capsys, root_dir, [*command_args, f"-i{did_id}"], assert_list)
    expected_calls: logType = {
        "sign_message_by_id": [(did_id, message.hex())],  # xck std
    }
    test_rpc_clients.wallet_rpc_client.check_log(expected_calls)


def test_nft_mint(capsys: object, get_test_cli_clients: tuple[TestRpcClients, Path]) -> None:
    test_rpc_clients, root_dir = get_test_cli_clients

    # set RPC Client
    class NFTCreateRpcClient(TestWalletRpcClient):
        async def get_nft_wallet_did(self, wallet_id: uint8) -> dict[str, Optional[str]]:
            self.add_to_log("get_nft_wallet_did", (wallet_id,))
            return {"did_id": "0xcee228b8638c67cb66a55085be99fa3b457ae5b56915896f581990f600b2c652"}

        async def mint_nft(
            self,
            wallet_id: int,
            royalty_address: Optional[str],
            target_address: Optional[str],
            hash: str,
            uris: list[str],
            meta_hash: str = "",
            meta_uris: Optional[list[str]] = None,
            license_hash: str = "",
            license_uris: Optional[list[str]] = None,
            edition_total: uint8 = uint8(1),
            edition_number: uint8 = uint8(1),
            fee: uint64 = uint64(0),
            royalty_percentage: int = 0,
            did_id: Optional[str] = None,
            reuse_puzhash: Optional[bool] = None,
            push: bool = True,
            timelock_info: ConditionValidTimes = ConditionValidTimes(),
        ) -> NFTMintNFTResponse:
            self.add_to_log(
                "mint_nft",
                (
                    wallet_id,
                    royalty_address,
                    target_address,
                    hash,
                    uris,
                    meta_hash,
                    meta_uris,
                    license_hash,
                    license_uris,
                    edition_total,
                    edition_number,
                    fee,
                    royalty_percentage,
                    did_id,
                    reuse_puzhash,
                    push,
                    timelock_info,
                ),
            )
            return NFTMintNFTResponse(
                [STD_UTX],
                [STD_TX],
                uint32(wallet_id),
                WalletSpendBundle([], G2Element()),
                bytes32.zeros.hex(),
            )

    inst_rpc_client = NFTCreateRpcClient()
    target_addr = encode_puzzle_hash(get_bytes32(2), "xck")
    test_rpc_clients.wallet_rpc_client = inst_rpc_client
    command_args = [
        "wallet",
        "nft",
        "mint",
        FINGERPRINT_ARG,
        "-i4",
        "--hash",
        "0x1234",
        "--uris",
        "https://example.com",
        "--target-address",
        target_addr,
        "-m0.5",
        "--reuse",
        "--valid-at",
        "100",
        "--expires-at",
        "150",
    ]
    # these are various things that should be in the output
    assert_list = [f"NFT minted Successfully with spend bundle: {STD_TX.spend_bundle}"]
    run_cli_command_and_assert(capsys, root_dir, command_args, assert_list)
    expected_calls: logType = {
        "get_nft_wallet_did": [(4,)],
        "mint_nft": [
            (
                4,
                None,
                "xck1qgpqyqszqgpqyqszqgpqyqszqgpqyqszqgpqyqszqgpqyqszqgpq29kamr",
                "0x1234",
                ["https://example.com"],
                TXConfig(
                    min_coin_amount=uint64(0),
                    max_coin_amount=uint64(18446744073709551615),
                    excluded_coin_amounts=[],
                    excluded_coin_ids=[],
                    reuse_puzhash=True,
                ),
                "",
                [],
                "",
                [],
                1,
                1,
                500000000000,
                0,
                "0xcee228b8638c67cb66a55085be99fa3b457ae5b56915896f581990f600b2c652",
                True,
                test_condition_valid_times,
            )
        ],
    }
    test_rpc_clients.wallet_rpc_client.check_log(expected_calls)


def test_nft_add_uri(capsys: object, get_test_cli_clients: tuple[TestRpcClients, Path]) -> None:
    test_rpc_clients, root_dir = get_test_cli_clients

    # set RPC Client
    class NFTAddUriRpcClient(TestWalletRpcClient):
        async def add_uri_to_nft(
            self,
            wallet_id: int,
            nft_coin_id: str,
            key: str,
            uri: str,
            fee: int,
            tx_config: TXConfig,
            push: bool,
            timelock_info: ConditionValidTimes = ConditionValidTimes(),
        ) -> NFTAddURIResponse:
            self.add_to_log("add_uri_to_nft", (wallet_id, nft_coin_id, key, uri, fee, tx_config, push, timelock_info))
            return NFTAddURIResponse([STD_UTX], [STD_TX], uint32(wallet_id), WalletSpendBundle([], G2Element()))

    inst_rpc_client = NFTAddUriRpcClient()
    nft_coin_id = get_bytes32(2).hex()
    test_rpc_clients.wallet_rpc_client = inst_rpc_client
    command_args = [
        "wallet",
        "nft",
        "add_uri",
        FINGERPRINT_ARG,
        "-i4",
        "--nft-coin-id",
        nft_coin_id,
        "--uri",
        "https://example.com/nft",
        "-m0.5",
        "--reuse",
        "--valid-at",
        "100",
        "--expires-at",
        "150",
    ]
    # these are various things that should be in the output
    assert STD_TX.spend_bundle is not None
    assert_list = [f"URI added successfully with spend bundle: {STD_TX.spend_bundle.to_json_dict()}"]
    run_cli_command_and_assert(capsys, root_dir, command_args, assert_list)
    expected_calls: logType = {
        "add_uri_to_nft": [
            (
                4,
                nft_coin_id,
                "u",
                "https://example.com/nft",
                500000000000,
                DEFAULT_TX_CONFIG.override(reuse_puzhash=True),
                True,
                test_condition_valid_times,
            )
        ],
    }
    test_rpc_clients.wallet_rpc_client.check_log(expected_calls)


def test_nft_transfer(capsys: object, get_test_cli_clients: tuple[TestRpcClients, Path]) -> None:
    test_rpc_clients, root_dir = get_test_cli_clients

    # set RPC Client
    class NFTTransferRpcClient(TestWalletRpcClient):
        async def transfer_nft(
            self,
            wallet_id: int,
            nft_coin_id: str,
            target_address: str,
            fee: int,
            tx_config: TXConfig,
            push: bool,
            timelock_info: ConditionValidTimes = ConditionValidTimes(),
        ) -> NFTTransferNFTResponse:
            self.add_to_log(
                "transfer_nft", (wallet_id, nft_coin_id, target_address, fee, tx_config, push, timelock_info)
            )
            return NFTTransferNFTResponse(
                [STD_UTX],
                [STD_TX],
                uint32(wallet_id),
                WalletSpendBundle([], G2Element()),
            )

    inst_rpc_client = NFTTransferRpcClient()
    nft_coin_id = get_bytes32(2).hex()
    target_address = encode_puzzle_hash(get_bytes32(2), "xck")
    test_rpc_clients.wallet_rpc_client = inst_rpc_client
    command_args = [
        "wallet",
        "nft",
        "transfer",
        FINGERPRINT_ARG,
        "-i4",
        "--nft-coin-id",
        nft_coin_id,
        "--target-address",
        target_address,
        "-m0.5",
        "--reuse",
        "--valid-at",
        "100",
        "--expires-at",
        "150",
    ]
    # these are various things that should be in the output
    assert STD_TX.spend_bundle is not None
    assert_list = ["NFT transferred successfully", f"spend bundle: {STD_TX.spend_bundle.to_json_dict()}"]
    run_cli_command_and_assert(capsys, root_dir, command_args, assert_list)
    expected_calls: logType = {
        "transfer_nft": [
            (
                4,
                nft_coin_id,
                target_address,
                500000000000,
                DEFAULT_TX_CONFIG.override(reuse_puzhash=True),
                True,
                test_condition_valid_times,
            )
        ],
    }
    test_rpc_clients.wallet_rpc_client.check_log(expected_calls)


def test_nft_list(capsys: object, get_test_cli_clients: tuple[TestRpcClients, Path]) -> None:
    test_rpc_clients, root_dir = get_test_cli_clients

    # set RPC Client
    class NFTListRpcClient(TestWalletRpcClient):
        async def list_nfts(self, wallet_id: int, num: int = 50, start_index: int = 0) -> dict[str, object]:
            self.add_to_log("list_nfts", (wallet_id, num, start_index))
            nft_list = []
            for i in range(start_index, start_index + num):
                index_bytes = bytes32([i] * 32)
                nft_list.append(
                    NFTInfo(
                        nft_id=encode_puzzle_hash(index_bytes, "nft"),
                        launcher_id=get_bytes32(1),
                        nft_coin_id=index_bytes,
                        nft_coin_confirmation_height=uint32(2),
                        owner_did=get_bytes32(2),
                        royalty_percentage=uint16(1000),
                        royalty_puzzle_hash=get_bytes32(3),
                        data_uris=["https://example.com/data"],
                        data_hash=bytes([4]),
                        metadata_uris=["https://example.com/mdata"],
                        metadata_hash=bytes([5]),
                        license_uris=["https://example.com/license"],
                        license_hash=bytes([6]),
                        edition_total=uint64(10),
                        edition_number=uint64(1),
                        updater_puzhash=get_bytes32(7),
                        chain_info="",
                        mint_height=uint32(1),
                        supports_did=True,
                        p2_address=get_bytes32(8),
                    ).to_json_dict()
                )
            return {"nft_list": nft_list}

    inst_rpc_client = NFTListRpcClient()
    launcher_ids = [bytes32([i] * 32).hex() for i in range(50, 60)]
    test_rpc_clients.wallet_rpc_client = inst_rpc_client
    command_args = [
        "wallet",
        "nft",
        "list",
        FINGERPRINT_ARG,
        "-i4",
        "--num",
        "10",
        "--start-index",
        "50",
    ]
    # these are various things that should be in the output
    assert_list = [
        "https://example.com/data",
        "did:chik:1qgpqyqszqgpqyqszqgpqyqszqgpqyqszqgpqyqszqgpqyqszqgpq4msw0c",
        *launcher_ids,
    ]
    run_cli_command_and_assert(capsys, root_dir, command_args, assert_list)
    expected_calls: logType = {
        "list_nfts": [(4, 10, 50)],
    }
    test_rpc_clients.wallet_rpc_client.check_log(expected_calls)


def test_nft_set_did(capsys: object, get_test_cli_clients: tuple[TestRpcClients, Path]) -> None:
    test_rpc_clients, root_dir = get_test_cli_clients

    # set RPC Client
    class NFTSetDidRpcClient(TestWalletRpcClient):
        async def set_nft_did(
            self,
            wallet_id: int,
            did_id: str,
            nft_coin_id: str,
            fee: int,
            tx_config: TXConfig,
            push: bool,
            timelock_info: ConditionValidTimes = ConditionValidTimes(),
        ) -> NFTSetNFTDIDResponse:
            self.add_to_log("set_nft_did", (wallet_id, did_id, nft_coin_id, fee, tx_config, push, timelock_info))
            return NFTSetNFTDIDResponse(
                [STD_UTX],
                [STD_TX],
                uint32(wallet_id),
                WalletSpendBundle([], G2Element()),
            )

    inst_rpc_client = NFTSetDidRpcClient()
    nft_coin_id = get_bytes32(2).hex()
    did_id = encode_puzzle_hash(get_bytes32(3), "did:chik:")
    test_rpc_clients.wallet_rpc_client = inst_rpc_client
    command_args = [
        "wallet",
        "nft",
        "set_did",
        FINGERPRINT_ARG,
        "-i4",
        "--nft-coin-id",
        nft_coin_id,
        "--did-id",
        did_id,
        "-m0.5",
        "--reuse",
        "--valid-at",
        "100",
        "--expires-at",
        "150",
    ]
    # these are various things that should be in the output
    assert STD_TX.spend_bundle is not None
    assert_list = [f"Transaction to set DID on NFT has been initiated with: {STD_TX.spend_bundle.to_json_dict()}"]
    run_cli_command_and_assert(capsys, root_dir, command_args, assert_list)
    expected_calls: logType = {
        "set_nft_did": [
            (
                4,
                did_id,
                nft_coin_id,
                500000000000,
                DEFAULT_TX_CONFIG.override(reuse_puzhash=True),
                True,
                test_condition_valid_times,
            )
        ],
    }
    test_rpc_clients.wallet_rpc_client.check_log(expected_calls)


def test_nft_get_info(capsys: object, get_test_cli_clients: tuple[TestRpcClients, Path]) -> None:
    test_rpc_clients, root_dir = get_test_cli_clients

    # set RPC Client

    inst_rpc_client = TestWalletRpcClient()
    nft_coin_id = get_bytes32(2).hex()
    test_rpc_clients.wallet_rpc_client = inst_rpc_client
    command_args = [
        "wallet",
        "nft",
        "get_info",
        FINGERPRINT_ARG,
        "--nft-coin-id",
        nft_coin_id,
    ]
    # these are various things that should be in the output
    assert_list = [
        f"Current NFT coin ID:       {nft_coin_id}",
        "Owner DID:                 did:chik:1qgpqyqszqgpqyqszqgpqyqszqgpqyqszqgpqyqszqgpqyqszqgpq4msw0c",
    ]
    run_cli_command_and_assert(capsys, root_dir, command_args, assert_list)
    expected_calls: logType = {
        "get_nft_info": [(nft_coin_id, True)],
    }
    test_rpc_clients.wallet_rpc_client.check_log(expected_calls)



from __future__ import annotations

import asyncio
import contextlib
from dataclasses import dataclass
from typing import TYPE_CHECKING, Callable, Optional

import aiosqlite
import pytest

# TODO: update after resolution in https://github.com/pytest-dev/pytest/issues/7469
from _pytest.fixtures import SubRequest

from chik._tests.util.db_connection import DBConnection, PathDBConnection
from chik._tests.util.misc import Marks, boolean_datacases, datacases
from chik.util.db_wrapper import DBWrapper2, ForeignKeyError, InternalError, NestedForeignKeyDelayedRequestError
from chik.util.task_referencer import create_referenced_task

if TYPE_CHECKING:
    ConnectionContextManager = contextlib.AbstractAsyncContextManager[aiosqlite.core.Connection]
    GetReaderMethod = Callable[[DBWrapper2], Callable[[], ConnectionContextManager]]


class UniqueError(Exception):
    """Used to uniquely trigger the exception path out of the context managers."""

    pass


async def increment_counter(db_wrapper: DBWrapper2) -> None:
    async with db_wrapper.writer_maybe_transaction() as connection:
        async with connection.execute("SELECT value FROM counter") as cursor:
            row = await cursor.fetchone()

        assert row is not None
        [old_value] = row

        await asyncio.sleep(0)

        new_value = old_value + 1
        await connection.execute("UPDATE counter SET value = :value", {"value": new_value})


async def decrement_counter(db_wrapper: DBWrapper2) -> None:
    async with db_wrapper.writer_maybe_transaction() as connection:
        async with connection.execute("SELECT value FROM counter") as cursor:
            row = await cursor.fetchone()

        assert row is not None
        [old_value] = row

        await asyncio.sleep(0)

        new_value = old_value - 1
        await connection.execute("UPDATE counter SET value = :value", {"value": new_value})


async def sum_counter(db_wrapper: DBWrapper2, output: list[int]) -> None:
    async with db_wrapper.reader_no_transaction() as connection:
        async with connection.execute("SELECT value FROM counter") as cursor:
            row = await cursor.fetchone()

        assert row is not None
        [value] = row

        output.append(value)


async def setup_table(db: DBWrapper2) -> None:
    async with db.writer_maybe_transaction() as conn:
        await conn.execute("CREATE TABLE counter(value INTEGER NOT NULL)")
        await conn.execute("INSERT INTO counter(value) VALUES(0)")


async def get_value(cursor: aiosqlite.Cursor) -> int:
    row = await cursor.fetchone()
    assert row
    return int(row[0])


async def query_value(connection: aiosqlite.Connection) -> int:
    async with connection.execute("SELECT value FROM counter") as cursor:
        return await get_value(cursor=cursor)


def _get_reader_no_transaction_method(db_wrapper: DBWrapper2) -> Callable[[], ConnectionContextManager]:
    return db_wrapper.reader_no_transaction


def _get_regular_reader_method(db_wrapper: DBWrapper2) -> Callable[[], ConnectionContextManager]:
    return db_wrapper.reader


@pytest.fixture(
    name="get_reader_method",
    params=[
        pytest.param(_get_reader_no_transaction_method, id="reader_no_transaction"),
        pytest.param(_get_regular_reader_method, id="reader"),
    ],
)
def get_reader_method_fixture(request: SubRequest) -> Callable[[], ConnectionContextManager]:
    # https://github.com/pytest-dev/pytest/issues/8763
    return request.param  # type: ignore[no-any-return]


@pytest.mark.anyio
@pytest.mark.parametrize(
    argnames="acquire_outside",
    argvalues=[pytest.param(False, id="not acquired outside"), pytest.param(True, id="acquired outside")],
)
async def test_concurrent_writers(acquire_outside: bool, get_reader_method: GetReaderMethod) -> None:
    async with DBConnection(2) as db_wrapper:
        await setup_table(db_wrapper)

        concurrent_task_count = 200

        async with contextlib.AsyncExitStack() as exit_stack:
            if acquire_outside:
                await exit_stack.enter_async_context(db_wrapper.writer_maybe_transaction())

            tasks = []
            for index in range(concurrent_task_count):
                task = create_referenced_task(increment_counter(db_wrapper))
                tasks.append(task)

        await asyncio.wait_for(asyncio.gather(*tasks), timeout=None)

        async with get_reader_method(db_wrapper)() as connection:
            async with connection.execute("SELECT value FROM counter") as cursor:
                row = await cursor.fetchone()

            assert row is not None
            [value] = row

    assert value == concurrent_task_count


@pytest.mark.anyio
async def test_writers_nests() -> None:
    async with DBConnection(2) as db_wrapper:
        await setup_table(db_wrapper)
        async with db_wrapper.writer_maybe_transaction() as conn1:
            async with conn1.execute("SELECT value FROM counter") as cursor:
                value = await get_value(cursor)
            async with db_wrapper.writer_maybe_transaction() as conn2:
                assert conn1 == conn2
                value += 1
                await conn2.execute("UPDATE counter SET value = :value", {"value": value})
                async with db_wrapper.writer_maybe_transaction() as conn3:
                    assert conn1 == conn3
                    async with conn3.execute("SELECT value FROM counter") as cursor:
                        value = await get_value(cursor)

    assert value == 1


@pytest.mark.anyio
async def test_writer_journal_mode_wal() -> None:
    async with PathDBConnection(2) as db_wrapper:
        async with db_wrapper.writer() as connection:
            async with connection.execute("PRAGMA journal_mode") as cursor:
                result = await cursor.fetchone()
                assert result == ("wal",)


@pytest.mark.anyio
async def test_reader_journal_mode_wal() -> None:
    async with PathDBConnection(2) as db_wrapper:
        async with db_wrapper.reader_no_transaction() as connection:
            async with connection.execute("PRAGMA journal_mode") as cursor:
                result = await cursor.fetchone()
                assert result == ("wal",)


@pytest.mark.anyio
async def test_partial_failure() -> None:
    values = []
    async with DBConnection(2) as db_wrapper:
        await setup_table(db_wrapper)
        async with db_wrapper.writer() as conn1:
            await conn1.execute("UPDATE counter SET value = 42")
            async with conn1.execute("SELECT value FROM counter") as cursor:
                values.append(await get_value(cursor))
            try:
                async with db_wrapper.writer() as conn2:
                    await conn2.execute("UPDATE counter SET value = 1337")
                    async with conn1.execute("SELECT value FROM counter") as cursor:
                        values.append(await get_value(cursor))
                    # this simulates a failure, which will cause a rollback of the
                    # write we just made, back to 42
                    raise RuntimeError("failure within a sub-transaction")
            except RuntimeError:
                # we expect to get here
                values.append(1)
            async with conn1.execute("SELECT value FROM counter") as cursor:
                values.append(await get_value(cursor))

    # the write of 1337 failed, and was restored to 42
    assert values == [42, 1337, 1, 42]


@pytest.mark.anyio
async def test_readers_nests(get_reader_method: GetReaderMethod) -> None:
    async with DBConnection(2) as db_wrapper:
        await setup_table(db_wrapper)

        async with get_reader_method(db_wrapper)() as conn1:
            async with get_reader_method(db_wrapper)() as conn2:
                assert conn1 == conn2
                async with get_reader_method(db_wrapper)() as conn3:
                    assert conn1 == conn3
                    async with conn3.execute("SELECT value FROM counter") as cursor:
                        value = await get_value(cursor)

    assert value == 0


@pytest.mark.anyio
async def test_readers_nests_writer(get_reader_method: GetReaderMethod) -> None:
    async with DBConnection(2) as db_wrapper:
        await setup_table(db_wrapper)

        async with db_wrapper.writer_maybe_transaction() as conn1:
            async with get_reader_method(db_wrapper)() as conn2:
                assert conn1 == conn2
                async with db_wrapper.writer_maybe_transaction() as conn3:
                    assert conn1 == conn3
                    async with conn3.execute("SELECT value FROM counter") as cursor:
                        value = await get_value(cursor)

    assert value == 0


@pytest.mark.parametrize(
    argnames="transactioned",
    argvalues=[
        pytest.param(True, id="transaction"),
        pytest.param(False, id="no transaction"),
    ],
)
@pytest.mark.anyio
async def test_only_transactioned_reader_ignores_writer(transactioned: bool) -> None:
    writer_committed = asyncio.Event()
    reader_read = asyncio.Event()

    async def write() -> None:
        try:
            async with db_wrapper.writer() as writer:
                assert reader is not writer

                await writer.execute("UPDATE counter SET value = 1")
        finally:
            writer_committed.set()

        await reader_read.wait()

        assert await query_value(connection=writer) == 1

    async with PathDBConnection(2) as db_wrapper:
        get_reader = db_wrapper.reader if transactioned else db_wrapper.reader_no_transaction

        await setup_table(db_wrapper)

        async with get_reader() as reader:
            assert await query_value(connection=reader) == 0

            task = create_referenced_task(write())
            await writer_committed.wait()

            assert await query_value(connection=reader) == 0 if transactioned else 1
            reader_read.set()

        await task

        async with get_reader() as reader:
            assert await query_value(connection=reader) == 1


@pytest.mark.anyio
async def test_reader_nests_and_ends_transaction() -> None:
    async with DBConnection(2) as db_wrapper:
        async with db_wrapper.reader() as reader:
            assert reader.in_transaction

            async with db_wrapper.reader() as inner_reader:
                assert inner_reader is reader
                assert reader.in_transaction

            assert reader.in_transaction

        assert not reader.in_transaction


@pytest.mark.anyio
async def test_writer_in_reader_works() -> None:
    async with PathDBConnection(2) as db_wrapper:
        await setup_table(db_wrapper)

        async with db_wrapper.reader() as reader:
            async with db_wrapper.writer() as writer:
                assert writer is not reader
                await writer.execute("UPDATE counter SET value = 1")
                assert await query_value(connection=writer) == 1
                assert await query_value(connection=reader) == 0

            assert await query_value(connection=reader) == 0


@pytest.mark.anyio
async def test_reader_transaction_is_deferred() -> None:
    async with DBConnection(2) as db_wrapper:
        await setup_table(db_wrapper)

        async with db_wrapper.reader() as reader:
            async with db_wrapper.writer() as writer:
                assert writer is not reader
                await writer.execute("UPDATE counter SET value = 1")
                assert await query_value(connection=writer) == 1

            # The deferred transaction initiation results in the transaction starting
            # here and thus reading the written value.
            assert await query_value(connection=reader) == 1


@pytest.mark.anyio
@pytest.mark.parametrize(
    argnames="acquire_outside",
    argvalues=[pytest.param(False, id="not acquired outside"), pytest.param(True, id="acquired outside")],
)
async def test_concurrent_readers(acquire_outside: bool, get_reader_method: GetReaderMethod) -> None:
    async with DBConnection(2) as db_wrapper:
        await setup_table(db_wrapper)

        async with db_wrapper.writer_maybe_transaction() as connection:
            await connection.execute("UPDATE counter SET value = 1")

        concurrent_task_count = 200

        async with contextlib.AsyncExitStack() as exit_stack:
            if acquire_outside:
                await exit_stack.enter_async_context(get_reader_method(db_wrapper)())

            tasks = []
            values: list[int] = []
            for index in range(concurrent_task_count):
                task = create_referenced_task(sum_counter(db_wrapper, values))
                tasks.append(task)

        await asyncio.wait_for(asyncio.gather(*tasks), timeout=None)

    assert values == [1] * concurrent_task_count


@pytest.mark.anyio
@pytest.mark.parametrize(
    argnames="acquire_outside",
    argvalues=[pytest.param(False, id="not acquired outside"), pytest.param(True, id="acquired outside")],
)
async def test_mixed_readers_writers(acquire_outside: bool, get_reader_method: GetReaderMethod) -> None:
    async with PathDBConnection(2) as db_wrapper:
        await setup_table(db_wrapper)

        async with db_wrapper.writer_maybe_transaction() as connection:
            await connection.execute("UPDATE counter SET value = 1")

        concurrent_task_count = 200

        async with contextlib.AsyncExitStack() as exit_stack:
            if acquire_outside:
                await exit_stack.enter_async_context(get_reader_method(db_wrapper)())

            tasks = []
            values: list[int] = []
            for index in range(concurrent_task_count):
                task = create_referenced_task(increment_counter(db_wrapper))
                tasks.append(task)
                task = create_referenced_task(decrement_counter(db_wrapper))
                tasks.append(task)
                task = create_referenced_task(sum_counter(db_wrapper, values))
                tasks.append(task)

        await asyncio.wait_for(asyncio.gather(*tasks), timeout=None)

        # we increment and decrement the counter an equal number of times. It should
        # end back at 1.
        async with get_reader_method(db_wrapper)() as connection:
            async with connection.execute("SELECT value FROM counter") as cursor:
                row = await cursor.fetchone()
                assert row is not None
                assert row[0] == 1

    # it's possible all increments or all decrements are run first
    assert len(values) == concurrent_task_count
    for v in values:
        assert v > -99
        assert v <= 100


@pytest.mark.parametrize(
    argnames=["manager_method", "expected"],
    argvalues=[
        [DBWrapper2.writer, True],
        [DBWrapper2.writer_maybe_transaction, True],
        [DBWrapper2.reader, True],
        [DBWrapper2.reader_no_transaction, False],
    ],
)
@pytest.mark.anyio
async def test_in_transaction_as_expected(
    manager_method: Callable[[DBWrapper2], ConnectionContextManager],
    expected: bool,
) -> None:
    async with DBConnection(2) as db_wrapper:
        await setup_table(db_wrapper)

        async with manager_method(db_wrapper) as connection:
            assert connection.in_transaction == expected


@pytest.mark.anyio
async def test_cancelled_reader_does_not_cancel_writer() -> None:
    async with DBConnection(2) as db_wrapper:
        await setup_table(db_wrapper)

        async with db_wrapper.writer() as writer:
            await writer.execute("UPDATE counter SET value = 1")

            with pytest.raises(UniqueError):
                async with db_wrapper.reader() as _:
                    raise UniqueError()

            assert await query_value(connection=writer) == 1

        assert await query_value(connection=writer) == 1


@boolean_datacases(name="initial", false="initially disabled", true="initially enabled")
@boolean_datacases(name="forced", false="forced disabled", true="forced enabled")
@pytest.mark.anyio
async def test_foreign_key_pragma_controlled_by_writer(initial: bool, forced: bool) -> None:
    async with DBConnection(2, foreign_keys=initial) as db_wrapper:
        async with db_wrapper.writer(foreign_key_enforcement_enabled=forced) as writer:
            async with writer.execute("PRAGMA foreign_keys") as cursor:
                result = await cursor.fetchone()
                assert result is not None
                [actual] = result

    assert actual == (1 if forced else 0)


@pytest.mark.anyio
async def test_foreign_key_pragma_rolls_back_on_foreign_key_error() -> None:
    async with DBConnection(2, foreign_keys=True, row_factory=aiosqlite.Row) as db_wrapper:
        async with db_wrapper.writer() as writer:
            async with writer.execute(
                """
                CREATE TABLE people(
                    id INTEGER NOT NULL,
                    friend INTEGER,
                    PRIMARY KEY (id),
                    FOREIGN KEY (friend) REFERENCES people
                )
                """
            ):
                pass

            async with writer.execute(
                "INSERT INTO people(id, friend) VALUES (:id, :friend)",
                {"id": 1, "friend": None},
            ):
                pass

            async with writer.execute(
                "INSERT INTO people(id, friend) VALUES (:id, :friend)",
                {"id": 2, "friend": 1},
            ):
                pass

        # make sure the writer raises a foreign key error on exit
        with pytest.raises(ForeignKeyError):
            async with db_wrapper.writer(foreign_key_enforcement_enabled=False) as writer:
                async with writer.execute("DELETE FROM people WHERE id = 1"):
                    pass

                # make sure a foreign key error can be detected here
                with pytest.raises(ForeignKeyError):
                    await db_wrapper._check_foreign_keys()

        async with writer.execute("SELECT * FROM people WHERE id = 1") as cursor:
            [person] = await cursor.fetchall()

        # make sure the delete was rolled back
        assert dict(person) == {"id": 1, "friend": None}


@dataclass
class RowFactoryCase:
    id: str
    factory: Optional[type[aiosqlite.Row]]
    marks: Marks = ()


row_factory_cases: list[RowFactoryCase] = [
    RowFactoryCase(id="default named tuple", factory=None),
    RowFactoryCase(id="aiosqlite row", factory=aiosqlite.Row),
]


@datacases(*row_factory_cases)
@pytest.mark.anyio
async def test_foreign_key_check_failure_error_message(case: RowFactoryCase) -> None:
    async with DBConnection(2, foreign_keys=True, row_factory=case.factory) as db_wrapper:
        async with db_wrapper.writer() as writer:
            async with writer.execute(
                """
                CREATE TABLE people(
                    id INTEGER NOT NULL,
                    friend INTEGER,
                    PRIMARY KEY (id),
                    FOREIGN KEY (friend) REFERENCES people
                )
                """
            ):
                pass

            async with writer.execute(
                "INSERT INTO people(id, friend) VALUES (:id, :friend)",
                {"id": 1, "friend": None},
            ):
                pass

            async with writer.execute(
                "INSERT INTO people(id, friend) VALUES (:id, :friend)",
                {"id": 2, "friend": 1},
            ):
                pass

        # make sure the writer raises a foreign key error on exit
        with pytest.raises(ForeignKeyError) as error:
            async with db_wrapper.writer(foreign_key_enforcement_enabled=False) as writer:
                async with writer.execute("DELETE FROM people WHERE id = 1"):
                    pass

        assert error.value.violations == [{"table": "people", "rowid": 2, "parent": "people", "fkid": 0}]


@pytest.mark.anyio
async def test_set_foreign_keys_fails_within_acquired_writer() -> None:
    async with DBConnection(2, foreign_keys=True) as db_wrapper:
        async with db_wrapper.writer():
            with pytest.raises(
                InternalError,
                match="Unable to set foreign key enforcement state while a writer is held",
            ):
                async with db_wrapper._set_foreign_key_enforcement(enabled=False):
                    pass  # pragma: no cover


@boolean_datacases(name="initial", false="initially disabled", true="initially enabled")
@pytest.mark.anyio
async def test_delayed_foreign_key_request_fails_when_nested(initial: bool) -> None:
    async with DBConnection(2, foreign_keys=initial) as db_wrapper:
        async with db_wrapper.writer():
            with pytest.raises(NestedForeignKeyDelayedRequestError):
                async with db_wrapper.writer(foreign_key_enforcement_enabled=True):
                    pass  # pragma: no cover



from __future__ import annotations

import math
from dataclasses import dataclass
from typing import Optional

from chik.types.fee_rate import FeeRate, FeeRateV2
from chik.util.ints import uint64
from chik.util.streamable import Streamable, streamable


@streamable
@dataclass(frozen=True)
class FeeEstimate(Streamable):
    """
    error: If error is not None, estimated_fee_rate is invalid, and `error` is a string describing the error.
    It can happen that only some requested FeeEstimates have errors, but others are valid.
    For example, an implementation may not have enough data yet for estimates farther in the future,
    or, an invalid parameter may have been passed.

    time_target: Epoch time in seconds we are targeting to include our `SpendBundle` in the blockchain.
    estimated_fee_rate: expressed in mojo per 1 klvm_cost. `estimated_fee` can be zero.
    """

    error: Optional[str]
    time_target: uint64  # unix time stamp in seconds
    estimated_fee_rate: FeeRate  # Mojos per klvm cost


@dataclass(frozen=True)
class FeeEstimateV2:
    error: Optional[str]
    time_target: uint64  # unix time stamp in seconds
    estimated_fee_rate: FeeRateV2  # Mojos per klvm cost


def fee_rate_v2_to_v1(fee_rate: FeeRateV2) -> FeeRate:
    return FeeRate(uint64(math.ceil(fee_rate.mojos_per_klvm_cost)))


def fee_estimate_v2_to_v1(fe: FeeEstimateV2) -> FeeEstimate:
    return FeeEstimate(fe.error, fe.time_target, FeeRate(uint64(math.ceil(fe.estimated_fee_rate.mojos_per_klvm_cost))))


@streamable
@dataclass(frozen=True)
class FeeEstimateGroup(Streamable):
    """
    If error is not None, at least one item in the list `estimates` is invalid.
    Estimates are expressed in mojos / 1 klvm_cost.
    """

    error: Optional[str]
    estimates: list[FeeEstimate]



from __future__ import annotations

from dataclasses import dataclass

import chik_rs

from chik.types.blockchain_format.reward_chain_block import RewardChainBlock
from chik.types.end_of_slot_bundle import EndOfSubSlotBundle
from chik.types.header_block import HeaderBlock
from chik.util.streamable import Streamable, streamable

SubEpochData = chik_rs.SubEpochData

# number of challenge blocks
# Average iters for challenge blocks
# |--A-R----R-------R--------R------R----R----------R-----R--R---|       Honest difficulty 1000
#           0.16

#  compute total reward chain blocks
# |----------------------------A---------------------------------|       Attackers chain 1000
#                            0.48
# total number of challenge blocks == total number of reward chain blocks


SubEpochChallengeSegment = chik_rs.SubEpochChallengeSegment
SubEpochSegments = chik_rs.SubEpochSegments
SubSlotData = chik_rs.SubSlotData


@streamable
@dataclass(frozen=True)
# this is used only for serialization to database
class RecentChainData(Streamable):
    recent_chain_data: list[HeaderBlock]


@streamable
@dataclass(frozen=True)
class ProofBlockHeader(Streamable):
    finished_sub_slots: list[EndOfSubSlotBundle]
    reward_chain_block: RewardChainBlock


@streamable
@dataclass(frozen=True)
class WeightProof(Streamable):
    sub_epochs: list[SubEpochData]
    sub_epoch_segments: list[SubEpochChallengeSegment]  # sampled sub epoch
    recent_chain_data: list[HeaderBlock]



from __future__ import annotations

from collections.abc import Iterable

from chik.protocols.shared_protocol import Capability
from chik.util.ints import uint16


def known_active_capabilities(values: Iterable[tuple[uint16, str]]) -> list[Capability]:
    # NOTE: order is not guaranteed
    # TODO: what if there's a claim for both supporting and not?
    #       presently it considers it supported
    filtered: set[Capability] = set()
    for value, state in values:
        if state != "1":
            continue

        try:
            filtered.add(Capability(value))
        except ValueError:
            pass

    # TODO: consider changing all uses to sets instead of lists
    return list(filtered)



from __future__ import annotations

from pathlib import Path
from typing import Optional

import click

from chik.cmds.cmd_classes import ChikCliContext
from chik.cmds.db_backup_func import db_backup_func
from chik.cmds.db_upgrade_func import db_upgrade_func
from chik.cmds.db_validate_func import db_validate_func


@click.group("db", help="Manage the blockchain database")
def db_cmd() -> None:
    pass


@db_cmd.command("upgrade", help="upgrade a v1 database to v2")
@click.option("--input", "in_db_path", default=None, type=click.Path(), help="specify input database file")
@click.option("--output", "out_db_path", default=None, type=click.Path(), help="specify output database file")
@click.option(
    "--no-update-config",
    default=False,
    is_flag=True,
    help="don't update config file to point to new database. When specifying a "
    "custom output file, the config will not be updated regardless",
)
@click.option(
    "--force",
    default=False,
    is_flag=True,
    help="force conversion despite warnings",
)
@click.pass_context
def db_upgrade_cmd(
    ctx: click.Context,
    in_db_path: Optional[str],
    out_db_path: Optional[str],
    no_update_config: bool,
    force: bool,
) -> None:
    try:
        db_upgrade_func(
            ChikCliContext.set_default(ctx).root_path,
            None if in_db_path is None else Path(in_db_path),
            None if out_db_path is None else Path(out_db_path),
            no_update_config=no_update_config,
            force=force,
        )
    except RuntimeError as e:
        print(f"FAILED: {e}")


@db_cmd.command("validate", help="validate the (v2) blockchain database. Does not verify proofs")
@click.option("--db", "in_db_path", default=None, type=click.Path(), help="Specifies which database file to validate")
@click.option(
    "--validate-blocks",
    default=False,
    is_flag=True,
    help="validate consistency of properties of the encoded blocks and block records",
)
@click.pass_context
def db_validate_cmd(ctx: click.Context, in_db_path: Optional[str], validate_blocks: bool) -> None:
    try:
        db_validate_func(
            ChikCliContext.set_default(ctx).root_path,
            None if in_db_path is None else Path(in_db_path),
            validate_blocks=validate_blocks,
        )
    except RuntimeError as e:
        print(f"FAILED: {e}")


@db_cmd.command("backup", help="backup the blockchain database using VACUUM INTO command")
@click.option("--backup_file", "db_backup_file", default=None, type=click.Path(), help="Specifies the backup file")
@click.option("--no_indexes", default=False, is_flag=True, help="Create backup without indexes")
@click.pass_context
def db_backup_cmd(ctx: click.Context, db_backup_file: Optional[str], no_indexes: bool) -> None:
    try:
        db_backup_func(
            ChikCliContext.set_default(ctx).root_path,
            None if db_backup_file is None else Path(db_backup_file),
            no_indexes=no_indexes,
        )
    except RuntimeError as e:
        print(f"FAILED: {e}")



# Package: utils

from __future__ import annotations

import asyncio
import logging
from typing import Optional


def cancel_task_safe(task: Optional[asyncio.Task[None]], log: Optional[logging.Logger] = None) -> None:
    if task is not None:
        try:
            task.cancel()
        except Exception as e:
            if log is not None:
                log.error(f"Error while canceling task.{e} {task}")



from __future__ import annotations

import ipaddress
from typing import TYPE_CHECKING, Any, ClassVar, Optional, cast

from chik.rpc.rpc_server import Endpoint, EndpointResult
from chik.seeder.crawler import Crawler
from chik.util.ws_message import WsRpcMessage, create_payload_dict


class CrawlerRpcApi:
    if TYPE_CHECKING:
        from chik.rpc.rpc_server import RpcApiProtocol

        _protocol_check: ClassVar[RpcApiProtocol] = cast("CrawlerRpcApi", None)

    def __init__(self, crawler: Crawler):
        self.service = crawler
        self.service_name = "chik_crawler"

    def get_routes(self) -> dict[str, Endpoint]:
        return {
            "/get_peer_counts": self.get_peer_counts,
            "/get_ips_after_timestamp": self.get_ips_after_timestamp,
        }

    async def _state_changed(self, change: str, change_data: Optional[dict[str, Any]] = None) -> list[WsRpcMessage]:
        payloads = []

        if change_data is None:
            change_data = await self.get_peer_counts({})

        if change in {"crawl_batch_completed", "loaded_initial_peers"}:
            payloads.append(create_payload_dict(change, change_data, self.service_name, "metrics"))

        return payloads

    async def get_peer_counts(self, _request: dict[str, Any]) -> EndpointResult:
        ipv6_addresses_count = 0
        for host in self.service.best_timestamp_per_peer.keys():
            try:
                ipaddress.IPv6Address(host)
                ipv6_addresses_count += 1
            except ipaddress.AddressValueError:
                continue

        reliable_peers = 0
        if self.service.crawl_store is not None:
            reliable_peers = self.service.crawl_store.get_reliable_peers()

        data = {
            "peer_counts": {
                "total_last_5_days": len(self.service.best_timestamp_per_peer),
                "reliable_nodes": reliable_peers,
                "ipv4_last_5_days": len(self.service.best_timestamp_per_peer) - ipv6_addresses_count,
                "ipv6_last_5_days": ipv6_addresses_count,
                "versions": self.service.versions,
            }
        }
        return data

    async def get_ips_after_timestamp(self, _request: dict[str, Any]) -> EndpointResult:
        after = _request.get("after", None)
        if after is None:
            raise ValueError("`after` is required and must be a unix timestamp")

        offset = _request.get("offset", 0)
        limit = _request.get("limit", 10000)

        matched_ips: list[str] = []
        for ip, timestamp in self.service.best_timestamp_per_peer.items():
            if timestamp > after:
                matched_ips.append(ip)

        matched_ips.sort()

        return {
            "ips": matched_ips[offset : (offset + limit)],
            "total": len(matched_ips),
        }



from __future__ import annotations

from typing import Any

from chik.util.db_wrapper import DBWrapper2


class KeyValStore:
    """
    Multipurpose persistent key-value store
    """

    db_wrapper: DBWrapper2

    @classmethod
    async def create(cls, db_wrapper: DBWrapper2):
        self = cls()
        self.db_wrapper = db_wrapper
        async with self.db_wrapper.writer_maybe_transaction() as conn:
            await conn.execute("CREATE TABLE IF NOT EXISTS key_val_store(key text PRIMARY KEY, value blob)")
            # Remove an old redundant index on the primary key
            # See https://github.com/Chik-Network/chik-blockchain/issues/10276
            await conn.execute("DROP INDEX IF EXISTS key_val_name")
        return self

    async def get_object(self, key: str, object_type: Any) -> Any:
        """
        Return bytes representation of stored object
        """

        async with self.db_wrapper.reader_no_transaction() as conn:
            cursor = await conn.execute("SELECT value from key_val_store WHERE key=?", (key,))
            row = await cursor.fetchone()
            await cursor.close()

        if row is None:
            return None

        return object_type.from_bytes(row[0])

    async def set_object(self, key: str, obj: Any):
        """
        Adds object to key val store. Obj MUST support __bytes__ and bytes() methods.
        """
        async with self.db_wrapper.writer_maybe_transaction() as conn:
            cursor = await conn.execute(
                "INSERT OR REPLACE INTO key_val_store VALUES(?, ?)",
                (key, obj.stream_to_bytes()),
            )
            await cursor.close()

    async def remove_object(self, key: str):
        async with self.db_wrapper.writer_maybe_transaction() as conn:
            cursor = await conn.execute("DELETE FROM key_val_store where key=?", (key,))
            await cursor.close()



from __future__ import annotations

import pytest

from chik._tests.util.setup_nodes import OldSimulatorsAndWallets
from chik.cmds.units import units
from chik.server.server import ChikServer
from chik.simulator.block_tools import BlockTools
from chik.simulator.full_node_simulator import FullNodeSimulator
from chik.types.peer_info import PeerInfo
from chik.util.ints import uint64
from chik.wallet.util.tx_config import DEFAULT_TX_CONFIG
from chik.wallet.wallet_node import WalletNode


@pytest.mark.anyio
@pytest.mark.parametrize(argnames="count", argvalues=[0, 1, 2, 5, 10])
@pytest.mark.parametrize(argnames="guarantee_transaction_blocks", argvalues=[False, True])
async def test_simulation_farm_blocks_to_puzzlehash(
    count: int,
    guarantee_transaction_blocks: bool,
    simulator_and_wallet: tuple[list[FullNodeSimulator], list[tuple[WalletNode, ChikServer]], BlockTools],
) -> None:
    [[full_node_api], _, _] = simulator_and_wallet

    # Starting at the beginning.
    assert full_node_api.full_node.blockchain.get_peak_height() is None

    await full_node_api.farm_blocks_to_puzzlehash(
        count=count, guarantee_transaction_blocks=guarantee_transaction_blocks
    )

    # The requested number of blocks had been processed.
    expected_height = None if count == 0 else count
    assert full_node_api.full_node.blockchain.get_peak_height() == expected_height


@pytest.mark.anyio
@pytest.mark.parametrize(argnames="count", argvalues=[0, 1, 2, 5, 10])
async def test_simulation_farm_blocks_to_wallet(
    count: int,
    simulator_and_wallet: tuple[list[FullNodeSimulator], list[tuple[WalletNode, ChikServer]], BlockTools],
) -> None:
    [[full_node_api], [[wallet_node, wallet_server]], _] = simulator_and_wallet

    await wallet_server.start_client(PeerInfo("127.0.0.1", full_node_api.server.get_port()), None)

    # Avoiding an attribute error below.
    assert wallet_node.wallet_state_manager is not None

    wallet = wallet_node.wallet_state_manager.main_wallet

    # Starting at the beginning.
    assert full_node_api.full_node.blockchain.get_peak_height() is None

    rewards = await full_node_api.farm_blocks_to_wallet(count=count, wallet=wallet)

    # The expected rewards have been received and confirmed.
    unconfirmed_balance = await wallet.get_unconfirmed_balance()
    confirmed_balance = await wallet.get_confirmed_balance()
    assert [unconfirmed_balance, confirmed_balance] == [rewards, rewards]


@pytest.mark.anyio
@pytest.mark.parametrize(
    argnames=["amount", "coin_count"],
    argvalues=[
        [0, 0],
        [1, 2],
        [(2 * units["chik"]) - 1, 2],
        [2 * units["chik"], 2],
        [(2 * units["chik"]) + 1, 4],
        [3 * units["chik"], 4],
        [10 * units["chik"], 10],
    ],
)
async def test_simulation_farm_rewards_to_wallet(
    amount: int,
    coin_count: int,
    simulator_and_wallet: tuple[list[FullNodeSimulator], list[tuple[WalletNode, ChikServer]], BlockTools],
) -> None:
    [[full_node_api], [[wallet_node, wallet_server]], _] = simulator_and_wallet

    await wallet_server.start_client(PeerInfo("127.0.0.1", full_node_api.server.get_port()), None)

    # Avoiding an attribute error below.
    assert wallet_node.wallet_state_manager is not None

    wallet = wallet_node.wallet_state_manager.main_wallet

    rewards = await full_node_api.farm_rewards_to_wallet(amount=amount, wallet=wallet)

    # At least the requested amount was farmed.
    assert rewards >= amount

    # The rewards amount is both received and confirmed.
    unconfirmed_balance = await wallet.get_unconfirmed_balance()
    confirmed_balance = await wallet.get_confirmed_balance()
    assert [unconfirmed_balance, confirmed_balance] == [rewards, rewards]

    # The expected number of coins were received.
    all_coin_records = await wallet.wallet_state_manager.coin_store.get_unspent_coins_for_wallet(wallet.id())
    assert len(all_coin_records) == coin_count


@pytest.mark.anyio
async def test_wait_transaction_records_entered_mempool(
    simulator_and_wallet: tuple[list[FullNodeSimulator], list[tuple[WalletNode, ChikServer]], BlockTools],
) -> None:
    repeats = 50
    tx_amount = 1
    [[full_node_api], [[wallet_node, wallet_server]], _] = simulator_and_wallet

    await wallet_server.start_client(PeerInfo("127.0.0.1", full_node_api.server.get_port()), None)

    # Avoiding an attribute hint issue below.
    assert wallet_node.wallet_state_manager is not None

    wallet = wallet_node.wallet_state_manager.main_wallet

    # generate some coins for repetitive testing
    await full_node_api.farm_rewards_to_wallet(amount=repeats * tx_amount, wallet=wallet)
    coins = await full_node_api.create_coins_with_amounts(amounts=[uint64(tx_amount)] * repeats, wallet=wallet)
    assert len(coins) == repeats

    # repeating just to try to expose any flakiness
    for coin in coins:
        async with wallet.wallet_state_manager.new_action_scope(DEFAULT_TX_CONFIG, push=True) as action_scope:
            await wallet.generate_signed_transaction(
                amount=uint64(tx_amount),
                puzzle_hash=await wallet_node.wallet_state_manager.main_wallet.get_new_puzzlehash(),
                action_scope=action_scope,
                coins={coin},
            )

        [tx] = action_scope.side_effects.transactions
        await full_node_api.wait_transaction_records_entered_mempool(records=action_scope.side_effects.transactions)
        assert tx.spend_bundle is not None
        assert full_node_api.full_node.mempool_manager.get_spendbundle(tx.spend_bundle.name()) is not None


@pytest.mark.anyio
async def test_process_transaction_records(
    simulator_and_wallet: tuple[list[FullNodeSimulator], list[tuple[WalletNode, ChikServer]], BlockTools],
) -> None:
    repeats = 50
    tx_amount = 1
    [[full_node_api], [[wallet_node, wallet_server]], _] = simulator_and_wallet

    await wallet_server.start_client(PeerInfo("127.0.0.1", full_node_api.server.get_port()), None)

    # Avoiding an attribute hint issue below.
    assert wallet_node.wallet_state_manager is not None

    wallet = wallet_node.wallet_state_manager.main_wallet

    # generate some coins for repetitive testing
    await full_node_api.farm_rewards_to_wallet(amount=repeats * tx_amount, wallet=wallet)
    coins = await full_node_api.create_coins_with_amounts(amounts=[uint64(tx_amount)] * repeats, wallet=wallet)
    assert len(coins) == repeats

    # repeating just to try to expose any flakiness
    for coin in coins:
        async with wallet.wallet_state_manager.new_action_scope(DEFAULT_TX_CONFIG, push=True) as action_scope:
            await wallet.generate_signed_transaction(
                amount=uint64(tx_amount),
                puzzle_hash=await wallet_node.wallet_state_manager.main_wallet.get_new_puzzlehash(),
                action_scope=action_scope,
                coins={coin},
            )

        await full_node_api.process_transaction_records(records=action_scope.side_effects.transactions)
        assert full_node_api.full_node.coin_store.get_coin_record(coin.name()) is not None


@pytest.mark.anyio
@pytest.mark.parametrize(
    argnames="amounts",
    argvalues=[
        *[pytest.param([uint64(1)] * n, id=f"1 mojo x {n}") for n in [0, 1, 10, 49, 51, 103]],
        *[
            pytest.param(list(uint64(x) for x in range(1, n + 1)), id=f"incrementing x {n}")
            for n in [1, 10, 49, 51, 103]
        ],
    ],
)
async def test_create_coins_with_amounts(
    self_hostname: str, amounts: list[uint64], simulator_and_wallet: OldSimulatorsAndWallets
) -> None:
    [[full_node_api], [[wallet_node, wallet_server]], _] = simulator_and_wallet
    await wallet_server.start_client(PeerInfo(self_hostname, full_node_api.server.get_port()), None)
    # Avoiding an attribute hint issue below.
    assert wallet_node.wallet_state_manager is not None
    wallet = wallet_node.wallet_state_manager.main_wallet
    await full_node_api.farm_rewards_to_wallet(amount=sum(amounts), wallet=wallet)
    # Get some more coins.  The creator helper doesn't get you all the coins you
    # need yet.
    await full_node_api.farm_blocks_to_wallet(count=2, wallet=wallet, timeout=30)
    coins = await full_node_api.create_coins_with_amounts(amounts=amounts, wallet=wallet, timeout=60)
    assert sorted(coin.amount for coin in coins) == sorted(amounts)


@pytest.mark.anyio
@pytest.mark.parametrize(
    argnames="amounts",
    argvalues=[
        [0],
        [5, -5],
        [4, 0],
    ],
    ids=lambda amounts: ", ".join(str(amount) for amount in amounts),
)
async def test_create_coins_with_invalid_amounts_raises(
    amounts: list[int],
    simulator_and_wallet: tuple[list[FullNodeSimulator], list[tuple[WalletNode, ChikServer]], BlockTools],
) -> None:
    [[full_node_api], [[wallet_node, wallet_server]], _] = simulator_and_wallet

    await wallet_server.start_client(PeerInfo("127.0.0.1", full_node_api.server.get_port()), None)

    # Avoiding an attribute hint issue below.backoff_times
    assert wallet_node.wallet_state_manager is not None

    wallet = wallet_node.wallet_state_manager.main_wallet

    with pytest.raises(Exception, match="Coins must have a positive value"):
        # Passing integers since the point is to test invalid values including
        # negatives that will not fit in a uint64.
        await full_node_api.create_coins_with_amounts(
            amounts=amounts,  # type: ignore[arg-type]
            wallet=wallet,
        )



from __future__ import annotations

import aiohttp
import pytest

from chik.daemon.server import WebSocketServer
from chik.simulator.block_tools import BlockTools
from chik.util.ws_message import create_payload


@pytest.mark.anyio
async def test_multiple_register_same(get_daemon: WebSocketServer, bt: BlockTools) -> None:
    ws_server = get_daemon
    config = bt.config

    daemon_port = config["daemon_port"]

    # setup receive service to connect to the daemon
    async with aiohttp.ClientSession() as client:
        async with client.ws_connect(
            f"wss://127.0.0.1:{daemon_port}",
            autoclose=True,
            autoping=True,
            ssl=bt.get_daemon_ssl_context(),
            max_msg_size=100 * 1024 * 1024,
        ) as ws:
            service_name = "test_service"
            data = {"service": service_name}
            payload = create_payload("register_service", data, service_name, "daemon")
            for _ in range(4):
                await ws.send_str(payload)
                await ws.receive()

            connections = ws_server.connections.get(service_name, set())
            assert len(connections) == 1


@pytest.mark.anyio
async def test_multiple_register_different(get_daemon: WebSocketServer, bt: BlockTools) -> None:
    ws_server = get_daemon
    config = bt.config

    daemon_port = config["daemon_port"]

    # setup receive service to connect to the daemon
    async with aiohttp.ClientSession() as client:
        async with client.ws_connect(
            f"wss://127.0.0.1:{daemon_port}",
            autoclose=True,
            autoping=True,
            ssl=bt.get_daemon_ssl_context(),
            max_msg_size=100 * 1024 * 1024,
        ) as ws:
            test_service_names = ["service1", "service2", "service3"]

            for service_name in test_service_names:
                data = {"service": service_name}
                payload = create_payload("register_service", data, service_name, "daemon")
                await ws.send_str(payload)
                await ws.receive()

            assert sorted(ws_server.connections.keys()) == test_service_names

            for service_name in test_service_names:
                connections = ws_server.connections.get(service_name, set())
                assert len(connections) == 1

        # Check after closing the connection
        for service_name in test_service_names:
            connections = ws_server.connections.get(service_name, set())
            assert connections == set()


@pytest.mark.anyio
async def test_remove_connection(get_daemon: WebSocketServer, bt: BlockTools) -> None:
    ws_server = get_daemon
    config = bt.config

    daemon_port = config["daemon_port"]

    # setup receive service to connect to the daemon
    async with aiohttp.ClientSession() as client:
        async with client.ws_connect(
            f"wss://127.0.0.1:{daemon_port}",
            autoclose=True,
            autoping=True,
            ssl=bt.get_daemon_ssl_context(),
            max_msg_size=100 * 1024 * 1024,
        ) as ws:
            test_service_names = ["service1", "service2", "service3", "service4", "service5"]

            for service_name in test_service_names:
                data = {"service": service_name}
                payload = create_payload("register_service", data, service_name, "daemon")
                await ws.send_str(payload)
                await ws.receive()

            assert sorted(ws_server.connections.keys()) == test_service_names

            connections = ws_server.connections.get(test_service_names[0], set())
            assert len(connections) == 1
            ws_to_remove = next(iter(connections))

            removed_names = ws_server.remove_connection(ws_to_remove)
            assert removed_names == test_service_names

            # remove again, should return empty list and not raise any errors
            removed_names = ws_server.remove_connection(ws_to_remove)
            assert removed_names == list()



from __future__ import annotations

import asyncio
from collections.abc import Sequence
from typing import Optional

import click

from chik.cmds import options
from chik.cmds.cmd_classes import ChikCliContext
from chik.cmds.cmds_util import CMDTXConfigLoader, tx_config_args, tx_out_cmd
from chik.cmds.param_types import AmountParamType, Bytes32ParamType, CliAmount, TransactionFeeParamType, Uint64ParamType
from chik.cmds.units import units
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.util.ints import uint64
from chik.wallet.conditions import ConditionValidTimes
from chik.wallet.transaction_record import TransactionRecord


@click.group("dao", short_help="Create, manage or show state of DAOs", no_args_is_help=True)
@click.pass_context
def dao_cmd(ctx: click.Context) -> None:
    pass


# ----------------------------------------------------------------------------------------
# ADD


@dao_cmd.command("add", short_help="Create a wallet for an existing DAO", no_args_is_help=True)
@click.option(
    "-wp",
    "--wallet-rpc-port",
    help="Set the port where the Wallet is hosting the RPC interface. See the rpc_port under wallet in config.yaml",
    type=int,
    default=None,
)
@click.option("-f", "--fingerprint", help="Set the fingerprint to specify which key to use", type=int)
@click.option("-n", "--name", help="Set the DAO wallet name", type=str)
@click.option(
    "-t",
    "--treasury-id",
    help="The Treasury ID of the DAO you want to track",
    type=Bytes32ParamType(),
    required=True,
)
@click.option(
    "-a",
    "--filter-amount",
    help="The minimum number of votes a proposal needs before the wallet will recognise it",
    type=Uint64ParamType(),
    default=uint64(1),
    show_default=True,
)
@click.pass_context
def dao_add_cmd(
    ctx: click.Context,
    wallet_rpc_port: Optional[int],
    fingerprint: int,
    treasury_id: bytes32,
    filter_amount: uint64,
    name: Optional[str],
) -> None:
    from chik.cmds.dao_funcs import add_dao_wallet

    asyncio.run(
        add_dao_wallet(
            ChikCliContext.set_default(ctx).root_path, wallet_rpc_port, fingerprint, name, treasury_id, filter_amount
        )
    )


# ----------------------------------------------------------------------------------------
# CREATE


@dao_cmd.command("create", short_help="Create a new DAO wallet and treasury", no_args_is_help=True)
@click.option(
    "-wp",
    "--wallet-rpc-port",
    help="Set the port where the Wallet is hosting the RPC interface. See the rpc_port under wallet in config.yaml",
    type=int,
    default=None,
)
@click.option("-f", "--fingerprint", help="Set the fingerprint to specify which key to use", type=int)
@click.option("-n", "--name", help="Set the DAO wallet name", type=str)
@click.option(
    "--proposal-timelock",
    help="The minimum number of blocks before a proposal can close",
    type=Uint64ParamType(),
    default="1000",
    show_default=True,
)
@click.option(
    "--soft-close",
    help="The number of blocks a proposal must remain unspent before closing",
    type=Uint64ParamType(),
    default="20",
    show_default=True,
)
@click.option(
    "--attendance-required",
    help="The minimum number of votes a proposal must receive to be accepted",
    type=Uint64ParamType(),
    required=True,
)
@click.option(
    "--pass-percentage",
    help="The percentage of 'yes' votes in basis points a proposal must receive to be accepted. 100% = 10000",
    type=Uint64ParamType(),
    default="5000",
    show_default=True,
)
@click.option(
    "--self-destruct",
    help="The number of blocks required before a proposal can be automatically removed",
    type=Uint64ParamType(),
    default="10000",
    show_default=True,
)
@click.option(
    "--oracle-delay",
    help="The number of blocks required between oracle spends of the treasury",
    type=Uint64ParamType(),
    default="50",
    show_default=True,
)
@click.option(
    "--proposal-minimum",
    help="The minimum amount (in xck) that a proposal must use to be created",
    type=AmountParamType(),
    default="1",
    show_default=True,
)
@click.option(
    "--filter-amount",
    help="The minimum number of votes a proposal needs before the wallet will recognise it",
    type=Uint64ParamType(),
    default="1",
    show_default=True,
)
@click.option(
    "--cat-amount",
    help="The number of DAO CATs (in mojos) to create when initializing the DAO",
    type=AmountParamType(),
    required=True,
)
@options.create_fee()
@click.option(
    "--fee-for-cat",
    help="Set the fees for the CAT creation transaction, in XCK.",
    type=TransactionFeeParamType(),
    default="0",
    show_default=True,
)
@tx_config_args
@tx_out_cmd()
@click.pass_context
def dao_create_cmd(
    ctx: click.Context,
    wallet_rpc_port: Optional[int],
    fingerprint: int,
    proposal_timelock: uint64,
    soft_close: uint64,
    attendance_required: uint64,
    pass_percentage: uint64,
    self_destruct: uint64,
    oracle_delay: uint64,
    proposal_minimum: CliAmount,
    filter_amount: uint64,
    cat_amount: CliAmount,
    name: Optional[str],
    fee: uint64,
    fee_for_cat: uint64,
    min_coin_amount: CliAmount,
    max_coin_amount: CliAmount,
    coins_to_exclude: Sequence[bytes32],
    amounts_to_exclude: Sequence[CliAmount],
    reuse: Optional[bool],
    push: bool,
    condition_valid_times: ConditionValidTimes,
) -> list[TransactionRecord]:
    from chik.cmds.dao_funcs import create_dao_wallet

    if self_destruct == proposal_timelock:
        raise ValueError("Self Destruct and Proposal Timelock cannot be the same value")

    print("Creating new DAO")

    return asyncio.run(
        create_dao_wallet(
            ChikCliContext.set_default(ctx).root_path,
            wallet_rpc_port,
            fingerprint,
            fee,
            fee_for_cat,
            name,
            proposal_timelock,
            soft_close,
            attendance_required,
            pass_percentage,
            self_destruct,
            oracle_delay,
            proposal_minimum.convert_amount(units["chik"]),
            filter_amount,
            cat_amount,
            CMDTXConfigLoader(
                min_coin_amount=min_coin_amount,
                max_coin_amount=max_coin_amount,
                excluded_coin_ids=list(coins_to_exclude),
                excluded_coin_amounts=list(amounts_to_exclude),
                reuse_puzhash=reuse,
            ),
            push,
            condition_valid_times=condition_valid_times,
        )
    )


# ----------------------------------------------------------------------------------------
# TREASURY INFO


@dao_cmd.command("get_id", short_help="Get the Treasury ID of a DAO", no_args_is_help=True)
@click.option(
    "-wp",
    "--wallet-rpc-port",
    help="Set the port where the Wallet is hosting the RPC interface. See the rpc_port under wallet in config.yaml",
    type=int,
    default=None,
)
@click.option("-f", "--fingerprint", help="Set the fingerprint to specify which key to use", type=int)
@click.option("-i", "--wallet-id", help="DAO Wallet ID", type=int, required=True)
@click.pass_context
def dao_get_id_cmd(
    ctx: click.Context,
    wallet_rpc_port: Optional[int],
    fingerprint: int,
    wallet_id: int,
) -> None:
    from chik.cmds.dao_funcs import get_treasury_id

    asyncio.run(get_treasury_id(ChikCliContext.set_default(ctx).root_path, wallet_rpc_port, fingerprint, wallet_id))


@dao_cmd.command("add_funds", short_help="Send funds to a DAO treasury", no_args_is_help=True)
@click.option(
    "-wp",
    "--wallet-rpc-port",
    help="Set the port where the Wallet is hosting the RPC interface. See the rpc_port under wallet in config.yaml",
    type=int,
    default=None,
)
@click.option("-f", "--fingerprint", help="Set the fingerprint to specify which key to use", type=int)
@click.option("-i", "--wallet-id", help="DAO Wallet ID which will receive the funds", type=int, required=True)
@click.option(
    "-w",
    "--funding-wallet-id",
    help="ID of the wallet to send funds from",
    type=int,
    required=True,
)
@click.option(
    "-a",
    "--amount",
    help="The amount of funds to send, in XCK or CATs",
    type=AmountParamType(),
    required=True,
)
@options.create_fee()
@tx_config_args
@tx_out_cmd()
@click.pass_context
def dao_add_funds_cmd(
    ctx: click.Context,
    wallet_rpc_port: Optional[int],
    fingerprint: int,
    wallet_id: int,
    funding_wallet_id: int,
    amount: CliAmount,
    fee: uint64,
    min_coin_amount: CliAmount,
    max_coin_amount: CliAmount,
    coins_to_exclude: Sequence[bytes32],
    amounts_to_exclude: Sequence[CliAmount],
    reuse: Optional[bool],
    push: bool,
    condition_valid_times: ConditionValidTimes,
) -> list[TransactionRecord]:
    from chik.cmds.dao_funcs import add_funds_to_treasury

    return asyncio.run(
        add_funds_to_treasury(
            ChikCliContext.set_default(ctx).root_path,
            wallet_rpc_port,
            fingerprint,
            wallet_id,
            funding_wallet_id,
            amount,
            fee,
            CMDTXConfigLoader(
                min_coin_amount=min_coin_amount,
                max_coin_amount=max_coin_amount,
                excluded_coin_ids=list(coins_to_exclude),
                excluded_coin_amounts=list(amounts_to_exclude),
                reuse_puzhash=reuse,
            ),
            push,
            condition_valid_times=condition_valid_times,
        )
    )


@dao_cmd.command("balance", short_help="Get the asset balances for a DAO treasury", no_args_is_help=True)
@click.option(
    "-wp",
    "--wallet-rpc-port",
    help="Set the port where the Wallet is hosting the RPC interface. See the rpc_port under wallet in config.yaml",
    type=int,
    default=None,
)
@click.option("-f", "--fingerprint", help="Set the fingerprint to specify which key to use", type=int)
@click.option("-i", "--wallet-id", help="Id of the wallet to use", type=int, required=True)
@click.pass_context
def dao_get_balance_cmd(
    ctx: click.Context,
    wallet_rpc_port: Optional[int],
    fingerprint: int,
    wallet_id: int,
) -> None:
    from chik.cmds.dao_funcs import get_treasury_balance

    asyncio.run(
        get_treasury_balance(ChikCliContext.set_default(ctx).root_path, wallet_rpc_port, fingerprint, wallet_id)
    )


@dao_cmd.command("rules", short_help="Get the current rules governing the DAO", no_args_is_help=True)
@click.option(
    "-wp",
    "--wallet-rpc-port",
    help="Set the port where the Wallet is hosting the RPC interface. See the rpc_port under wallet in config.yaml",
    type=int,
    default=None,
)
@click.option("-f", "--fingerprint", help="Set the fingerprint to specify which key to use", type=int)
@click.option("-i", "--wallet-id", help="Id of the wallet to use", type=int, required=True)
@click.pass_context
def dao_rules_cmd(
    ctx: click.Context,
    wallet_rpc_port: Optional[int],
    fingerprint: int,
    wallet_id: int,
) -> None:
    from chik.cmds.dao_funcs import get_rules

    asyncio.run(get_rules(ChikCliContext.set_default(ctx).root_path, wallet_rpc_port, fingerprint, wallet_id))


# ----------------------------------------------------------------------------------------
# LIST/SHOW PROPOSALS


@dao_cmd.command("list_proposals", short_help="List proposals for the DAO", no_args_is_help=True)
@click.option(
    "-wp",
    "--wallet-rpc-port",
    help="Set the port where the Wallet is hosting the RPC interface. See the rpc_port under wallet in config.yaml",
    type=int,
    default=None,
)
@click.option("-f", "--fingerprint", help="Set the fingerprint to specify which key to use", type=int)
@click.option("-i", "--wallet-id", help="Id of the wallet to use", type=int, required=True)
@click.option(
    "-c",
    "--include-closed",
    help="Include previously closed proposals",
    is_flag=True,
)
@click.pass_context
def dao_list_proposals_cmd(
    ctx: click.Context,
    wallet_rpc_port: Optional[int],
    fingerprint: int,
    wallet_id: int,
    include_closed: Optional[bool],
) -> None:
    from chik.cmds.dao_funcs import list_proposals

    if not include_closed:
        include_closed = False

    asyncio.run(
        list_proposals(
            ChikCliContext.set_default(ctx).root_path, wallet_rpc_port, fingerprint, wallet_id, include_closed
        )
    )


@dao_cmd.command("show_proposal", short_help="Show the details of a specific proposal", no_args_is_help=True)
@click.option(
    "-wp",
    "--wallet-rpc-port",
    help="Set the port where the Wallet is hosting the RPC interface. See the rpc_port under wallet in config.yaml",
    type=int,
    default=None,
)
@click.option("-f", "--fingerprint", help="Set the fingerprint to specify which key to use", type=int)
@click.option("-i", "--wallet-id", help="Id of the wallet to use", type=int, required=True)
@click.option(
    "-p",
    "--proposal_id",
    help="The ID of the proposal to fetch",
    type=str,
    required=True,
)
@click.pass_context
def dao_show_proposal_cmd(
    ctx: click.Context,
    wallet_rpc_port: Optional[int],
    fingerprint: int,
    wallet_id: int,
    proposal_id: str,
) -> None:
    from chik.cmds.dao_funcs import show_proposal

    asyncio.run(
        show_proposal(ChikCliContext.set_default(ctx).root_path, wallet_rpc_port, fingerprint, wallet_id, proposal_id)
    )


# ----------------------------------------------------------------------------------------
# VOTE


@dao_cmd.command("vote", short_help="Vote on a DAO proposal", no_args_is_help=True)
@click.option(
    "-wp",
    "--wallet-rpc-port",
    help="Set the port where the Wallet is hosting the RPC interface. See the rpc_port under wallet in config.yaml",
    type=int,
    default=None,
)
@click.option("-f", "--fingerprint", help="Set the fingerprint to specify which key to use", type=int)
@click.option("-i", "--wallet-id", help="Id of the wallet to use", type=int, required=True)
@click.option(
    "-p",
    "--proposal-id",
    help="The ID of the proposal you are voting on",
    type=str,
    required=True,
)
@click.option(
    "-a",
    "--vote-amount",
    help="The number of votes you want to cast",
    type=Uint64ParamType(),
    required=True,
)
@click.option(
    "-n",
    "--vote-no",
    help="Use this option to vote against a proposal. If not present then the vote is for the proposal",
    is_flag=True,
)
@options.create_fee()
@tx_config_args
@tx_out_cmd()
@click.pass_context
def dao_vote_cmd(
    ctx: click.Context,
    wallet_rpc_port: Optional[int],
    fingerprint: int,
    wallet_id: int,
    proposal_id: str,
    vote_amount: uint64,
    vote_no: Optional[bool],
    fee: uint64,
    min_coin_amount: CliAmount,
    max_coin_amount: CliAmount,
    coins_to_exclude: Sequence[bytes32],
    amounts_to_exclude: Sequence[CliAmount],
    reuse: Optional[bool],
    push: bool,
    condition_valid_times: ConditionValidTimes,
) -> list[TransactionRecord]:
    from chik.cmds.dao_funcs import vote_on_proposal

    is_yes_vote = False if vote_no else True

    return asyncio.run(
        vote_on_proposal(
            ChikCliContext.set_default(ctx).root_path,
            wallet_rpc_port,
            fingerprint,
            wallet_id,
            proposal_id,
            vote_amount,
            is_yes_vote,
            fee,
            CMDTXConfigLoader(
                min_coin_amount=min_coin_amount,
                max_coin_amount=max_coin_amount,
                excluded_coin_ids=list(coins_to_exclude),
                excluded_coin_amounts=list(amounts_to_exclude),
                reuse_puzhash=reuse,
            ),
            push,
            condition_valid_times=condition_valid_times,
        )
    )


# ----------------------------------------------------------------------------------------
# CLOSE PROPOSALS


@dao_cmd.command("close_proposal", short_help="Close a DAO proposal", no_args_is_help=True)
@click.option(
    "-wp",
    "--wallet-rpc-port",
    help="Set the port where the Wallet is hosting the RPC interface. See the rpc_port under wallet in config.yaml",
    type=int,
    default=None,
)
@click.option("-f", "--fingerprint", help="Set the fingerprint to specify which key to use", type=int)
@click.option("-i", "--wallet-id", help="Id of the wallet to use", type=int, required=True)
@click.option(
    "-p",
    "--proposal-id",
    help="The ID of the proposal you are voting on",
    type=str,
    required=True,
)
@click.option(
    "-d",
    "--self-destruct",
    help="If a proposal is broken, use self destruct to force it to close",
    is_flag=True,
    default=False,
)
@options.create_fee()
@tx_config_args
@tx_out_cmd()
@click.pass_context
def dao_close_proposal_cmd(
    ctx: click.Context,
    wallet_rpc_port: Optional[int],
    fingerprint: int,
    wallet_id: int,
    proposal_id: str,
    self_destruct: bool,
    fee: uint64,
    min_coin_amount: CliAmount,
    max_coin_amount: CliAmount,
    coins_to_exclude: Sequence[bytes32],
    amounts_to_exclude: Sequence[CliAmount],
    reuse: Optional[bool],
    push: bool,
    condition_valid_times: ConditionValidTimes,
) -> list[TransactionRecord]:
    from chik.cmds.dao_funcs import close_proposal

    return asyncio.run(
        close_proposal(
            ChikCliContext.set_default(ctx).root_path,
            wallet_rpc_port,
            fingerprint,
            wallet_id,
            fee,
            proposal_id,
            self_destruct,
            CMDTXConfigLoader(
                min_coin_amount=min_coin_amount,
                max_coin_amount=max_coin_amount,
                excluded_coin_ids=list(coins_to_exclude),
                excluded_coin_amounts=list(amounts_to_exclude),
                reuse_puzhash=reuse,
            ),
            push,
            condition_valid_times=condition_valid_times,
        )
    )


# ----------------------------------------------------------------------------------------
# LOCKUP COINS


@dao_cmd.command("lockup_coins", short_help="Lock DAO CATs for voting", no_args_is_help=True)
@click.option(
    "-wp",
    "--wallet-rpc-port",
    help="Set the port where the Wallet is hosting the RPC interface. See the rpc_port under wallet in config.yaml",
    type=int,
    default=None,
)
@click.option("-f", "--fingerprint", help="Set the fingerprint to specify which key to use", type=int)
@click.option("-i", "--wallet-id", help="Id of the wallet to use", type=int, required=True)
@click.option(
    "-a",
    "--amount",
    help="The amount of CATs (not mojos) to lock in voting mode",
    type=AmountParamType(),
    required=True,
)
@options.create_fee()
@tx_config_args
@tx_out_cmd()
@click.pass_context
def dao_lockup_coins_cmd(
    ctx: click.Context,
    wallet_rpc_port: Optional[int],
    fingerprint: int,
    wallet_id: int,
    amount: CliAmount,
    fee: uint64,
    min_coin_amount: CliAmount,
    max_coin_amount: CliAmount,
    coins_to_exclude: Sequence[bytes32],
    amounts_to_exclude: Sequence[CliAmount],
    reuse: Optional[bool],
    push: bool,
    condition_valid_times: ConditionValidTimes,
) -> list[TransactionRecord]:
    from chik.cmds.dao_funcs import lockup_coins

    return asyncio.run(
        lockup_coins(
            ChikCliContext.set_default(ctx).root_path,
            wallet_rpc_port,
            fingerprint,
            wallet_id,
            amount,
            fee,
            CMDTXConfigLoader(
                min_coin_amount=min_coin_amount,
                max_coin_amount=max_coin_amount,
                excluded_coin_ids=list(coins_to_exclude),
                excluded_coin_amounts=list(amounts_to_exclude),
                reuse_puzhash=reuse,
            ),
            push,
            condition_valid_times=condition_valid_times,
        )
    )


@dao_cmd.command("release_coins", short_help="Release closed proposals from DAO CATs", no_args_is_help=True)
@click.option(
    "-wp",
    "--wallet-rpc-port",
    help="Set the port where the Wallet is hosting the RPC interface. See the rpc_port under wallet in config.yaml",
    type=int,
    default=None,
)
@click.option("-f", "--fingerprint", help="Set the fingerprint to specify which key to use", type=int)
@click.option("-i", "--wallet-id", help="Id of the wallet to use", type=int, required=True)
@options.create_fee()
@tx_config_args
@tx_out_cmd()
@click.pass_context
def dao_release_coins_cmd(
    ctx: click.Context,
    wallet_rpc_port: Optional[int],
    fingerprint: int,
    wallet_id: int,
    fee: uint64,
    min_coin_amount: CliAmount,
    max_coin_amount: CliAmount,
    coins_to_exclude: Sequence[bytes32],
    amounts_to_exclude: Sequence[CliAmount],
    reuse: Optional[bool],
    push: bool,
    condition_valid_times: ConditionValidTimes,
) -> list[TransactionRecord]:
    from chik.cmds.dao_funcs import release_coins

    return asyncio.run(
        release_coins(
            ChikCliContext.set_default(ctx).root_path,
            wallet_rpc_port,
            fingerprint,
            wallet_id,
            fee,
            CMDTXConfigLoader(
                min_coin_amount=min_coin_amount,
                max_coin_amount=max_coin_amount,
                excluded_coin_ids=list(coins_to_exclude),
                excluded_coin_amounts=list(amounts_to_exclude),
                reuse_puzhash=reuse,
            ),
            push,
            condition_valid_times=condition_valid_times,
        )
    )


@dao_cmd.command("exit_lockup", short_help="Release DAO CATs from voting mode", no_args_is_help=True)
@click.option(
    "-wp",
    "--wallet-rpc-port",
    help="Set the port where the Wallet is hosting the RPC interface. See the rpc_port under wallet in config.yaml",
    type=int,
    default=None,
)
@click.option("-f", "--fingerprint", help="Set the fingerprint to specify which key to use", type=int)
@click.option("-i", "--wallet-id", help="Id of the wallet to use", type=int, required=True)
@options.create_fee()
@tx_config_args
@tx_out_cmd()
@click.pass_context
def dao_exit_lockup_cmd(
    ctx: click.Context,
    wallet_rpc_port: Optional[int],
    fingerprint: int,
    wallet_id: int,
    fee: uint64,
    min_coin_amount: CliAmount,
    max_coin_amount: CliAmount,
    coins_to_exclude: Sequence[bytes32],
    amounts_to_exclude: Sequence[CliAmount],
    reuse: Optional[bool],
    push: bool,
    condition_valid_times: ConditionValidTimes,
) -> list[TransactionRecord]:
    from chik.cmds.dao_funcs import exit_lockup

    return asyncio.run(
        exit_lockup(
            ChikCliContext.set_default(ctx).root_path,
            wallet_rpc_port,
            fingerprint,
            wallet_id,
            fee,
            CMDTXConfigLoader(
                min_coin_amount=min_coin_amount,
                max_coin_amount=max_coin_amount,
                excluded_coin_ids=list(coins_to_exclude),
                excluded_coin_amounts=list(amounts_to_exclude),
                reuse_puzhash=reuse,
            ),
            push,
            condition_valid_times=condition_valid_times,
        )
    )


# ----------------------------------------------------------------------------------------
# CREATE PROPOSALS


@dao_cmd.group("create_proposal", short_help="Create and add a proposal to a DAO", no_args_is_help=True)
@click.pass_context
def dao_proposal(ctx: click.Context) -> None:
    pass


@dao_proposal.command("spend", short_help="Create a proposal to spend DAO funds", no_args_is_help=True)
@click.option(
    "-wp",
    "--wallet-rpc-port",
    help="Set the port where the Wallet is hosting the RPC interface. See the rpc_port under wallet in config.yaml",
    type=int,
    default=None,
)
@click.option("-f", "--fingerprint", help="Set the fingerprint to specify which key to use", type=int)
@click.option("-i", "--wallet-id", help="Id of the wallet to use", type=int, required=True)
@click.option(
    "-t",
    "--to-address",
    help="The address the proposal will send funds to",
    type=str,
    required=False,
    default=None,
)
@click.option(
    "-a",
    "--amount",
    help="The amount of funds the proposal will send (in mojos)",
    type=str,
    required=False,
    default=None,
)
@click.option(
    "-v",
    "--vote-amount",
    help="The number of votes to add",
    type=int,
    required=False,
    default=None,
)
@click.option(
    "--asset-id",
    help="The asset id of the funds the proposal will send. Leave blank for xck",
    type=str,
    required=False,
    default=None,
)
@click.option(
    "-j",
    "--from-json",
    help="Path to a json file containing a list of additions, for use in proposals with multiple spends",
    type=str,
    required=False,
    default=None,
)
@options.create_fee()
@tx_config_args
@tx_out_cmd()
@click.pass_context
def dao_create_spend_proposal_cmd(
    ctx: click.Context,
    wallet_rpc_port: Optional[int],
    fingerprint: int,
    wallet_id: int,
    fee: uint64,
    vote_amount: Optional[int],
    to_address: Optional[str],
    amount: Optional[str],
    asset_id: Optional[str],
    from_json: Optional[str],
    min_coin_amount: CliAmount,
    max_coin_amount: CliAmount,
    coins_to_exclude: Sequence[bytes32],
    amounts_to_exclude: Sequence[CliAmount],
    reuse: Optional[bool],
    push: bool,
    condition_valid_times: ConditionValidTimes,
) -> list[TransactionRecord]:
    from chik.cmds.dao_funcs import create_spend_proposal

    return asyncio.run(
        create_spend_proposal(
            ChikCliContext.set_default(ctx).root_path,
            wallet_rpc_port,
            fingerprint,
            wallet_id,
            fee,
            vote_amount,
            to_address,
            amount,
            asset_id,
            from_json,
            CMDTXConfigLoader(
                min_coin_amount=min_coin_amount,
                max_coin_amount=max_coin_amount,
                excluded_coin_ids=list(coins_to_exclude),
                excluded_coin_amounts=list(amounts_to_exclude),
                reuse_puzhash=reuse,
            ),
            push,
            condition_valid_times=condition_valid_times,
        )
    )


@dao_proposal.command("update", short_help="Create a proposal to change the DAO rules", no_args_is_help=True)
@click.option(
    "-wp",
    "--wallet-rpc-port",
    help="Set the port where the Wallet is hosting the RPC interface. See the rpc_port under wallet in config.yaml",
    type=int,
    default=None,
)
@click.option("-f", "--fingerprint", help="Set the fingerprint to specify which key to use", type=int)
@click.option("-i", "--wallet-id", help="Id of the wallet to use", type=int, required=True)
@click.option(
    "-v",
    "--vote-amount",
    help="The number of votes to add",
    type=Uint64ParamType(),
    required=False,
    default=None,
)
@click.option(
    "--proposal-timelock",
    help="The new minimum number of blocks before a proposal can close",
    type=Uint64ParamType(),
    default=None,
    required=False,
)
@click.option(
    "--soft-close",
    help="The number of blocks a proposal must remain unspent before closing",
    type=Uint64ParamType(),
    default=None,
    required=False,
)
@click.option(
    "--attendance-required",
    help="The minimum number of votes a proposal must receive to be accepted",
    type=Uint64ParamType(),
    default=None,
    required=False,
)
@click.option(
    "--pass-percentage",
    help="The percentage of 'yes' votes in basis points a proposal must receive to be accepted. 100% = 10000",
    type=Uint64ParamType(),
    default=None,
    required=False,
)
@click.option(
    "--self-destruct",
    help="The number of blocks required before a proposal can be automatically removed",
    type=Uint64ParamType(),
    default=None,
    required=False,
)
@click.option(
    "--oracle-delay",
    help="The number of blocks required between oracle spends of the treasury",
    type=Uint64ParamType(),
    default=None,
    required=False,
)
@options.create_fee()
@tx_config_args
@tx_out_cmd()
@click.pass_context
def dao_create_update_proposal_cmd(
    ctx: click.Context,
    wallet_rpc_port: Optional[int],
    fingerprint: int,
    wallet_id: int,
    fee: uint64,
    vote_amount: Optional[uint64],
    proposal_timelock: Optional[uint64],
    soft_close: Optional[uint64],
    attendance_required: Optional[uint64],
    pass_percentage: Optional[uint64],
    self_destruct: Optional[uint64],
    oracle_delay: Optional[uint64],
    min_coin_amount: CliAmount,
    max_coin_amount: CliAmount,
    coins_to_exclude: Sequence[bytes32],
    amounts_to_exclude: Sequence[CliAmount],
    reuse: Optional[bool],
    push: bool,
    condition_valid_times: ConditionValidTimes,
) -> list[TransactionRecord]:
    from chik.cmds.dao_funcs import create_update_proposal

    return asyncio.run(
        create_update_proposal(
            ChikCliContext.set_default(ctx).root_path,
            wallet_rpc_port,
            fingerprint,
            wallet_id,
            fee,
            vote_amount,
            proposal_timelock,
            soft_close,
            attendance_required,
            pass_percentage,
            self_destruct,
            oracle_delay,
            CMDTXConfigLoader(
                min_coin_amount=min_coin_amount,
                max_coin_amount=max_coin_amount,
                excluded_coin_ids=list(coins_to_exclude),
                excluded_coin_amounts=list(amounts_to_exclude),
                reuse_puzhash=reuse,
            ),
            push,
            condition_valid_times=condition_valid_times,
        )
    )


@dao_proposal.command("mint", short_help="Create a proposal to mint new DAO CATs", no_args_is_help=True)
@click.option(
    "-wp",
    "--wallet-rpc-port",
    help="Set the port where the Wallet is hosting the RPC interface. See the rpc_port under wallet in config.yaml",
    type=int,
    default=None,
)
@click.option("-f", "--fingerprint", help="Set the fingerprint to specify which key to use", type=int)
@click.option("-i", "--wallet-id", help="Id of the wallet to use", type=int, required=True)
@click.option(
    "-a",
    "--amount",
    help="The amount of new cats the proposal will mint (in mojos)",
    type=Uint64ParamType(),
    required=True,
)
@click.option(
    "-t",
    "--to-address",
    help="The address new cats will be minted to",
    type=str,
    required=True,
    default=None,
)
@click.option(
    "-v",
    "--vote-amount",
    help="The number of votes to add",
    type=int,
    required=False,
    default=None,
)
@options.create_fee()
@tx_config_args
@tx_out_cmd()
@click.pass_context
def dao_create_mint_proposal_cmd(
    ctx: click.Context,
    wallet_rpc_port: Optional[int],
    fingerprint: int,
    wallet_id: int,
    fee: uint64,
    amount: uint64,
    to_address: str,
    vote_amount: Optional[int],
    min_coin_amount: CliAmount,
    max_coin_amount: CliAmount,
    coins_to_exclude: Sequence[bytes32],
    amounts_to_exclude: Sequence[CliAmount],
    reuse: Optional[bool],
    push: bool,
    condition_valid_times: ConditionValidTimes,
) -> list[TransactionRecord]:
    from chik.cmds.dao_funcs import create_mint_proposal

    return asyncio.run(
        create_mint_proposal(
            ChikCliContext.set_default(ctx).root_path,
            wallet_rpc_port,
            fingerprint,
            wallet_id,
            fee,
            amount,
            to_address,
            vote_amount,
            CMDTXConfigLoader(
                min_coin_amount=min_coin_amount,
                max_coin_amount=max_coin_amount,
                excluded_coin_ids=list(coins_to_exclude),
                excluded_coin_amounts=list(amounts_to_exclude),
                reuse_puzhash=reuse,
            ),
            push,
            condition_valid_times=condition_valid_times,
        )
    )


# ----------------------------------------------------------------------------------------

dao_cmd.add_command(dao_add_cmd)
dao_cmd.add_command(dao_create_cmd)
dao_cmd.add_command(dao_add_funds_cmd)
dao_cmd.add_command(dao_get_balance_cmd)
dao_cmd.add_command(dao_list_proposals_cmd)
dao_cmd.add_command(dao_show_proposal_cmd)
dao_cmd.add_command(dao_vote_cmd)
dao_cmd.add_command(dao_close_proposal_cmd)
dao_cmd.add_command(dao_lockup_coins_cmd)
dao_cmd.add_command(dao_exit_lockup_cmd)
dao_cmd.add_command(dao_release_coins_cmd)
dao_cmd.add_command(dao_proposal)



from __future__ import annotations

from typing import Optional

import pytest

from chik._tests.blockchain.blockchain_test_utils import _validate_and_add_block
from chik._tests.util.blockchain import create_blockchain
from chik._tests.util.time_out_assert import time_out_assert
from chik.consensus.block_record import BlockRecord
from chik.consensus.blockchain import Blockchain
from chik.consensus.difficulty_adjustment import get_next_sub_slot_iters_and_difficulty
from chik.consensus.make_sub_epoch_summary import next_sub_epoch_summary
from chik.protocols import timelord_protocol
from chik.server.server import ChikServer
from chik.simulator.block_tools import BlockTools
from chik.simulator.full_node_simulator import FullNodeSimulator
from chik.timelord.timelord_api import TimelordAPI
from chik.types.aliases import FullNodeService
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.types.blockchain_format.sub_epoch_summary import SubEpochSummary
from chik.types.full_block import FullBlock
from chik.types.unfinished_block import UnfinishedBlock
from chik.util.ints import uint128


class TestNewPeak:
    @pytest.mark.anyio
    async def test_timelord_new_peak_basic(
        self, bt: BlockTools, timelord: tuple[TimelordAPI, ChikServer], default_1000_blocks: list[FullBlock]
    ) -> None:
        async with create_blockchain(bt.constants, 2) as (b1, _):
            async with create_blockchain(bt.constants, 2) as (b2, _):
                timelord_api, _ = timelord
                for block in default_1000_blocks:
                    await _validate_and_add_block(b1, block)
                    await _validate_and_add_block(b2, block)

                peak = timelord_peak_from_block(b1, default_1000_blocks[-1])
                assert peak is not None
                assert timelord_api.timelord.new_peak is None
                await timelord_api.new_peak_timelord(peak)
                assert timelord_api.timelord.new_peak is not None
                assert timelord_api.timelord.new_peak.reward_chain_block.height == peak.reward_chain_block.height
                blocks = bt.get_consecutive_blocks(1, default_1000_blocks)
                await _validate_and_add_block(b1, blocks[-1])
                await _validate_and_add_block(b2, blocks[-1])

                await timelord_api.new_peak_timelord(timelord_peak_from_block(b1, blocks[-1]))
                assert timelord_api.timelord.new_peak.reward_chain_block.height == blocks[-1].height

                blocks_1 = bt.get_consecutive_blocks(2, blocks)
                await _validate_and_add_block(b1, blocks_1[-2])
                await _validate_and_add_block(b1, blocks_1[-1])
                await timelord_api.new_peak_timelord(timelord_peak_from_block(b1, blocks_1[-2]))
                await timelord_api.new_peak_timelord(timelord_peak_from_block(b1, blocks_1[-1]))
                assert timelord_api.timelord.new_peak.reward_chain_block.height == blocks_1[-1].height

                # new unknown peak, weight less then curr peak
                blocks_2 = bt.get_consecutive_blocks(1, blocks)
                await _validate_and_add_block(b2, blocks_2[-1])
                await timelord_api.new_peak_timelord(timelord_peak_from_block(b2, blocks_2[-1]))
                assert timelord_api.timelord.last_state.last_weight == blocks_1[-1].weight
                assert timelord_api.timelord.last_state.total_iters == blocks_1[-1].reward_chain_block.total_iters

    @pytest.mark.anyio
    async def test_timelord_new_peak_unfinished_not_orphaned(
        self, bt: BlockTools, timelord: tuple[TimelordAPI, ChikServer], default_1000_blocks: list[FullBlock]
    ) -> None:
        async with create_blockchain(bt.constants, 2) as (b1, _):
            timelord_api, _ = timelord
            for block in default_1000_blocks:
                await _validate_and_add_block(b1, block)

            peak = timelord_peak_from_block(b1, default_1000_blocks[-1])
            assert peak is not None
            assert timelord_api.timelord.new_peak is None
            await timelord_api.new_peak_timelord(peak)
            assert timelord_api.timelord.new_peak is not None
            assert timelord_api.timelord.new_peak.reward_chain_block.get_hash() == peak.reward_chain_block.get_hash()

            # make two new blocks on tip
            blocks_1 = bt.get_consecutive_blocks(2, default_1000_blocks)
            block_1 = blocks_1[-2]
            block_2 = blocks_1[-1]
            await _validate_and_add_block(b1, block_1)
            await _validate_and_add_block(b1, block_2)

            block_record = b1.block_record(block_2.header_hash)
            sub_slot_iters, difficulty = get_next_sub_slot_iters_and_difficulty(
                bt.constants, len(block_1.finished_sub_slots) > 0, b1.block_record(block_1.prev_header_hash), b1
            )

            timelord_unf_block = timelord_protocol.NewUnfinishedBlockTimelord(
                block_1.reward_chain_block.get_unfinished(),
                difficulty,
                sub_slot_iters,
                block_1.foliage,
                next_sub_epoch_summary(bt.constants, b1, block_record.required_iters, block_1, True),
                await get_rc_prev(b1, block_1),
            )
            await timelord_api.new_unfinished_block_timelord(timelord_unf_block)

            assert timelord_api.timelord.unfinished_blocks[-1].get_hash() == timelord_unf_block.get_hash()
            new_peak = timelord_peak_from_block(b1, block_2)
            assert timelord_unf_block.reward_chain_block.total_iters <= new_peak.reward_chain_block.total_iters
            await timelord_api.new_peak_timelord(new_peak)

            await time_out_assert(60, peak_new_peak_is_none, True, timelord_api)

            assert (
                timelord_api.timelord.last_state.peak.reward_chain_block.get_hash()
                == new_peak.reward_chain_block.get_hash()
            )

    @pytest.mark.anyio
    async def test_timelord_new_peak_unfinished_orphaned(
        self,
        one_node: tuple[list[FullNodeService], list[FullNodeSimulator], BlockTools],
        timelord: tuple[TimelordAPI, ChikServer],
        default_1000_blocks: list[FullBlock],
    ) -> None:
        [full_node_service], _, bt = one_node
        full_node = full_node_service._node
        async with create_blockchain(bt.constants, 2) as (b1, _):
            async with create_blockchain(bt.constants, 2) as (b2, _):
                timelord_api, _ = timelord
                for block in default_1000_blocks:
                    await _validate_and_add_block(b1, block)
                    await _validate_and_add_block(b2, block)
                    await full_node.add_block(block)

                peak = timelord_peak_from_block(b1, default_1000_blocks[-1])
                assert peak is not None
                assert timelord_api.timelord.new_peak is None
                await timelord_api.new_peak_timelord(peak)
                assert timelord_api.timelord.new_peak is not None
                assert (
                    timelord_api.timelord.new_peak.reward_chain_block.get_hash() == peak.reward_chain_block.get_hash()
                )

                # make two new blocks on tip, block_2 has higher total iterations
                block_1 = bt.get_consecutive_blocks(1, default_1000_blocks)[-1]
                block_2 = bt.get_consecutive_blocks(
                    1, default_1000_blocks, min_signage_point=block_1.reward_chain_block.signage_point_index
                )[-1]

                # make sure block_2 has higher iterations then block_1
                assert block_2.total_iters > block_1.total_iters
                # make sure block_1 and block_2 have higher iterations then peak
                assert block_1.total_iters > default_1000_blocks[-1].total_iters

                await _validate_and_add_block(b1, block_1)
                await _validate_and_add_block(b2, block_2)

                block_record_1 = b1.block_record(block_1.header_hash)
                sub_slot_iters, difficulty = get_next_sub_slot_iters_and_difficulty(
                    bt.constants,
                    len(block_1.finished_sub_slots) > 0,
                    b1.block_record(block_1.prev_header_hash),
                    b1,
                )

                timelord_unf_block = timelord_protocol.NewUnfinishedBlockTimelord(
                    block_1.reward_chain_block.get_unfinished(),
                    difficulty,
                    sub_slot_iters,
                    block_1.foliage,
                    next_sub_epoch_summary(bt.constants, b1, block_record_1.required_iters, block_1, True),
                    await get_rc_prev(b1, block_1),
                )
                await timelord_api.new_unfinished_block_timelord(timelord_unf_block)
                assert timelord_api.timelord.unfinished_blocks[-1].get_hash() == timelord_unf_block.get_hash()
                new_peak = timelord_peak_from_block(b2, block_2)

                # timelord knows unfinished block_1 that has lower iterations,
                # add block_2 peak and make sure we skip it and prefer to finish block_1
                assert timelord_unf_block.reward_chain_block.total_iters <= new_peak.reward_chain_block.total_iters
                await timelord_api.new_peak_timelord(new_peak)
                await time_out_assert(60, peak_new_peak_is_none, True, timelord_api)

                # check that peak did not change
                assert (
                    timelord_api.timelord.last_state.peak.reward_chain_block.get_hash()
                    == peak.reward_chain_block.get_hash()
                )
                # check unfinished block_1 is still in cache
                assert timelord_api.timelord.unfinished_blocks[-1].get_hash() == timelord_unf_block.get_hash()

                # full node gets block_1 unfinished
                block_1_unf = UnfinishedBlock(
                    block_1.finished_sub_slots,
                    block_1.reward_chain_block.get_unfinished(),
                    block_1.challenge_chain_sp_proof,
                    block_1.reward_chain_sp_proof,
                    block_1.foliage,
                    block_1.foliage_transaction_block,
                    block_1.transactions_info,
                    block_1.transactions_generator,
                    [],
                )
                await full_node.add_unfinished_block(block_1_unf, None)
                unf: UnfinishedBlock = full_node.full_node_store.get_unfinished_block(block_1_unf.partial_hash)
                assert unf.get_hash() == block_1_unf.get_hash()
                # full node peak is block_2
                await full_node.add_block(block_2)
                curr = await full_node.blockchain.get_full_peak()
                assert block_2.header_hash == curr.header_hash

                # full_node gets finished block_1
                response = timelord_protocol.NewInfusionPointVDF(
                    block_1_unf.partial_hash,
                    block_1.reward_chain_block.challenge_chain_ip_vdf,
                    block_1.challenge_chain_ip_proof,
                    block_1.reward_chain_block.reward_chain_ip_vdf,
                    block_1.reward_chain_ip_proof,
                    block_1.reward_chain_block.infused_challenge_chain_ip_vdf,
                    block_1.infused_challenge_chain_ip_proof,
                )

                await full_node.new_infusion_point_vdf(response)
                peak_after_unf_infusion = await full_node.blockchain.get_full_peak()
                # assert full node switched peak to block_1 since it has the same height as block_2 but lower iterations
                assert peak_after_unf_infusion.header_hash == block_1.header_hash

    @pytest.mark.anyio
    async def test_timelord_new_peak_unfinished_orphaned_overflow(
        self, bt: BlockTools, timelord: tuple[TimelordAPI, ChikServer], default_1000_blocks: list[FullBlock]
    ) -> None:
        async with create_blockchain(bt.constants, 2) as (b1, _):
            async with create_blockchain(bt.constants, 2) as (b2, _):
                timelord_api, _ = timelord
                for block in default_1000_blocks:
                    await _validate_and_add_block(b1, block)
                    await _validate_and_add_block(b2, block)

                peak = timelord_peak_from_block(b1, default_1000_blocks[-1])
                assert peak is not None
                assert timelord_api.timelord.new_peak is None
                await timelord_api.new_peak_timelord(peak)
                assert timelord_api.timelord.new_peak is not None
                assert (
                    timelord_api.timelord.new_peak.reward_chain_block.get_hash() == peak.reward_chain_block.get_hash()
                )

                # make two new blocks on tip
                block_1 = bt.get_consecutive_blocks(1, default_1000_blocks, time_per_block=9, force_overflow=True)[-1]
                block_2 = bt.get_consecutive_blocks(
                    1, default_1000_blocks, seed=b"data", time_per_block=50, skip_slots=1
                )[-1]
                # make sure block_2 has higher iterations
                assert block_2.total_iters >= block_1.total_iters

                await _validate_and_add_block(b1, block_1)
                await _validate_and_add_block(b2, block_2)

                block_record = b1.block_record(block_1.header_hash)
                sub_slot_iters, difficulty = get_next_sub_slot_iters_and_difficulty(
                    bt.constants,
                    len(block_1.finished_sub_slots) > 0,
                    b1.block_record(block_1.prev_header_hash),
                    b1,
                )

                timelord_unf_block = timelord_protocol.NewUnfinishedBlockTimelord(
                    block_1.reward_chain_block.get_unfinished(),
                    difficulty,
                    sub_slot_iters,
                    block_1.foliage,
                    next_sub_epoch_summary(bt.constants, b1, block_record.required_iters, block_1, True),
                    await get_rc_prev(b1, block_1),
                )
                await timelord_api.new_unfinished_block_timelord(timelord_unf_block)

                assert timelord_api.timelord.overflow_blocks[-1].get_hash() == timelord_unf_block.get_hash()
                new_peak = timelord_peak_from_block(b2, block_2)
                assert timelord_unf_block.reward_chain_block.total_iters <= new_peak.reward_chain_block.total_iters
                await timelord_api.new_peak_timelord(new_peak)

                await time_out_assert(60, peak_new_peak_is_none, True, timelord_api)

                assert (
                    timelord_api.timelord.last_state.peak.reward_chain_block.get_hash()
                    == peak.reward_chain_block.get_hash()
                )

    @pytest.mark.anyio
    async def test_timelord_new_peak_unfinished_eos(
        self, bt: BlockTools, timelord: tuple[TimelordAPI, ChikServer], default_1000_blocks: list[FullBlock]
    ) -> None:
        async with create_blockchain(bt.constants, 2) as (b1, _):
            async with create_blockchain(bt.constants, 2) as (b2, _):
                timelord_api, _ = timelord
                for block in default_1000_blocks:
                    await _validate_and_add_block(b1, block)
                    await _validate_and_add_block(b2, block)

                peak = timelord_peak_from_block(b1, default_1000_blocks[-1])
                assert peak is not None
                assert timelord_api.timelord.new_peak is None
                await timelord_api.new_peak_timelord(peak)
                assert timelord_api.timelord.new_peak is not None
                assert (
                    timelord_api.timelord.new_peak.reward_chain_block.get_hash() == peak.reward_chain_block.get_hash()
                )

                # make two new blocks on tip, block_2 is in a new slot
                block_1 = bt.get_consecutive_blocks(1, default_1000_blocks)[-1]
                block_2 = bt.get_consecutive_blocks(
                    1, default_1000_blocks, skip_slots=1, skip_overflow=True, seed=b"data"
                )[-1]

                # make sure block_2 has higher iterations
                assert block_2.total_iters >= block_1.total_iters

                await _validate_and_add_block(b1, block_1)
                await _validate_and_add_block(b2, block_2)

                block_record = b2.block_record(block_2.header_hash)
                sub_slot_iters, difficulty = get_next_sub_slot_iters_and_difficulty(
                    bt.constants,
                    len(block_2.finished_sub_slots) > 0,
                    b1.block_record(block_2.prev_header_hash),
                    b1,
                )

                timelord_unf_block = timelord_protocol.NewUnfinishedBlockTimelord(
                    block_2.reward_chain_block.get_unfinished(),
                    difficulty,
                    sub_slot_iters,
                    block_2.foliage,
                    next_sub_epoch_summary(bt.constants, b1, block_record.required_iters, block_2, True),
                    await get_rc_prev(b2, block_2),
                )
                timelord_api.timelord.last_state.set_state(block_2.finished_sub_slots[-1])

                # add unfinished and make sure we cache it
                await timelord_api.new_unfinished_block_timelord(timelord_unf_block)
                assert timelord_api.timelord.unfinished_blocks[-1].get_hash() == timelord_unf_block.get_hash()
                new_peak = timelord_peak_from_block(b1, block_1)
                assert timelord_unf_block.reward_chain_block.total_iters >= new_peak.reward_chain_block.total_iters
                await timelord_api.new_peak_timelord(new_peak)
                await time_out_assert(60, peak_new_peak_is_none, True, timelord_api)

                # make sure we switch to lower iteration peak
                assert (
                    timelord_api.timelord.last_state.peak.reward_chain_block.get_hash()
                    == new_peak.reward_chain_block.get_hash()
                )


async def get_rc_prev(blockchain: Blockchain, block: FullBlock) -> bytes32:
    if block.reward_chain_block.signage_point_index == 0:
        # find first in slot and find slot challenge
        blk = blockchain.block_record(block.header_hash)
        while blk.first_in_sub_slot is False:
            blk = blockchain.block_record(block.prev_header_hash)
        full_blk = await blockchain.get_full_block(blk.header_hash)
        assert full_blk is not None
        sub_slot = None
        for s in full_blk.finished_sub_slots:
            if s is not None and s.challenge_chain.get_hash() == block.reward_chain_block.pos_ss_cc_challenge_hash:
                sub_slot = s
        if sub_slot is None:
            assert block.reward_chain_block.pos_ss_cc_challenge_hash == blockchain.constants.GENESIS_CHALLENGE
            rc_prev = blockchain.constants.GENESIS_CHALLENGE
        else:
            rc_prev = sub_slot.reward_chain.get_hash()
    else:
        assert block.reward_chain_block.reward_chain_sp_vdf is not None
        rc_prev = block.reward_chain_block.reward_chain_sp_vdf.challenge
    return rc_prev


def get_recent_reward_challenges(blockchain: Blockchain) -> list[tuple[bytes32, uint128]]:
    peak = blockchain.get_peak()
    if peak is None:
        return []
    recent_rc: list[tuple[bytes32, uint128]] = []
    curr: Optional[BlockRecord] = peak
    while curr is not None and len(recent_rc) < 2 * blockchain.constants.MAX_SUB_SLOT_BLOCKS:
        if curr != peak:
            recent_rc.append((curr.reward_infusion_new_challenge, curr.total_iters))
        if curr.first_in_sub_slot:
            assert curr.finished_reward_slot_hashes is not None
            sub_slot_total_iters = curr.ip_sub_slot_total_iters(blockchain.constants)
            # Start from the most recent
            for rc in reversed(curr.finished_reward_slot_hashes):
                if sub_slot_total_iters < curr.sub_slot_iters:
                    break
                recent_rc.append((rc, sub_slot_total_iters))
                sub_slot_total_iters = uint128(sub_slot_total_iters - curr.sub_slot_iters)
        curr = blockchain.try_block_record(curr.prev_hash)
    return list(reversed(recent_rc))


def timelord_peak_from_block(
    blockchain: Blockchain,
    block: FullBlock,
) -> timelord_protocol.NewPeakTimelord:
    peak = blockchain.block_record(block.header_hash)
    _, difficulty = get_next_sub_slot_iters_and_difficulty(blockchain.constants, False, peak, blockchain)
    ses: Optional[SubEpochSummary] = next_sub_epoch_summary(
        blockchain.constants, blockchain, peak.required_iters, block, True
    )

    curr = peak
    while not curr.is_challenge_block(blockchain.constants) and not curr.first_in_sub_slot:
        curr = blockchain.block_record(curr.prev_hash)

    if curr.is_challenge_block(blockchain.constants):
        last_csb_or_eos = curr.total_iters
    else:
        last_csb_or_eos = curr.ip_sub_slot_total_iters(blockchain.constants)

    curr = peak
    passed_ses_height_but_not_yet_included = True
    while (curr.height % blockchain.constants.SUB_EPOCH_BLOCKS) != 0:
        if curr.sub_epoch_summary_included:
            passed_ses_height_but_not_yet_included = False
        curr = blockchain.block_record(curr.prev_hash)
    if curr.sub_epoch_summary_included or curr.height == 0:
        passed_ses_height_but_not_yet_included = False
    return timelord_protocol.NewPeakTimelord(
        block.reward_chain_block,
        difficulty,
        peak.deficit,
        peak.sub_slot_iters,
        ses,
        get_recent_reward_challenges(blockchain),
        last_csb_or_eos,
        passed_ses_height_but_not_yet_included,
    )


def peak_new_peak_is_none(timelord: TimelordAPI) -> bool:
    return timelord.timelord.new_peak is None



from __future__ import annotations

from dataclasses import dataclass

from chik.types.condition_opcodes import ConditionOpcode


@dataclass(frozen=True)
class ConditionWithArgs:
    """
    This structure is used to store parsed KLVM conditions
    Conditions in KLVM have either format of (opcode, var1) or (opcode, var1, var2)
    """

    opcode: ConditionOpcode
    vars: list[bytes]
